# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pathlib import Path
from typing import List, Optional
from mxnet.gluon import nn
from mxnet.gluon.trainer import Trainer
from .base import Callback


class ModelSaverCallback(Callback):  # type: ignore
    """
    The model saver callback saves the model during training at exponential
    frequency.

    Attributes:
        network: The network that was trained. Not available prior to training.
        saved_parameters: The parameters saved for the different milestones. Should only be
            accessed after training has finished and should not be modified.
        training_times: The training times in seconds for the different milestones.
        num_gradient_updates: The number of gradient updates for the different milestones.
    """

    def __init__(
        self,
        directory: Path,
        milestones: List[float],
    ):
        """
        Args:
            directory: The directory into which the model parameters ought to be saved. Models are
                stored as `model_<seq>.params` where `<seq>` starts with 0 and reaches
                `count - 1`. Note that the directory must exist.
            milestones: The number of seconds after which the model should be saved.
        """
        assert all(
            x < y for x, y in zip(milestones, milestones[1:])
        ), "Time milestones must be increasing."

        super().__init__()
        self.directory = directory
        self.seq = 0
        self.batch_count = 0
        self.network: Optional[nn.HybridBlock] = None
        self.saved_parameters: List[Path] = []
        self.training_times: List[float] = []
        self.num_gradient_updates: List[int] = []
        self.milestones = milestones

    def on_train_start(self, trainer: Trainer) -> None:
        self.seq = 0
        self.batch_count = 0
        self.network = None
        self.saved_parameters = []
        self.training_times = []
        self.num_gradient_updates = []

    def on_network_initialization_end(self, network: nn.HybridBlock) -> None:
        self.network = network

    def on_train_batch_end(
        self, network: nn.HybridBlock, time_elapsed: float
    ) -> None:
        self.batch_count += 1
        if (
            len(self.milestones) > self.seq
            and time_elapsed > self.milestones[self.seq]
        ):
            file = self.directory / f"model_{self.seq}.params"
            network.save_parameters(file.absolute().as_posix())
            self.saved_parameters.append(file)
            self.training_times.append(time_elapsed)
            self.num_gradient_updates.append(self.batch_count)
            self.seq += 1



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Any, Callable, Optional
import torch
from torchmetrics import Metric


class NormalizedDeviation(Metric):
    # pylint: disable=arguments-differ

    def __init__(
        self,
        rescale: bool = False,
        compute_on_step: bool = True,
        dist_sync_on_step: bool = False,
        process_group: Optional[Any] = None,
        dist_sync_fn: Callable = None,
    ):
        super().__init__(
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )
        self.rescale = rescale
        self.add_state(
            "numerator", default=torch.as_tensor(0.0), dist_reduce_fx="sum"
        )
        self.add_state(
            "denom", default=torch.as_tensor(0.0), dist_reduce_fx="sum"
        )

    def update(
        self,
        y_pred: torch.Tensor,
        y_true: torch.Tensor,
        mask: torch.Tensor,
        scales: Optional[torch.Tensor] = None,
    ) -> None:
        if self.rescale:
            m = scales[:, 0].unsqueeze(1)
            std = scales[:, 1].unsqueeze(1)
            y_true = y_true * std + m
            y_pred = y_pred * std.unsqueeze(2) + m.unsqueeze(2)

        idx_median = y_pred.size()[-1] // 2
        y_pred = y_pred[:, : y_true.size()[1], idx_median]
        losses = (y_pred - y_true).abs()
        losses = torch.mul(losses, mask)

        self.numerator += losses.sum()
        self.denom += y_true.abs().sum()

    def compute(self) -> torch.Tensor:
        return self.numerator / self.denom



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from functools import partial
from inspect import isfunction

import numpy as np

import torch
from torch import nn, einsum
import torch.nn.functional as F


def default(val, d):
    if val is not None:
        return val
    return d() if isfunction(d) else d


def extract(a, t, x_shape):
    b, *_ = t.shape
    out = a.gather(-1, t)
    return out.reshape(b, *((1,) * (len(x_shape) - 1)))


def noise_like(shape, device, repeat=False):
    repeat_noise = lambda: torch.randn((1, *shape[1:]), device=device).repeat(
        shape[0], *((1,) * (len(shape) - 1))
    )
    noise = lambda: torch.randn(shape, device=device)
    return repeat_noise() if repeat else noise()


def cosine_beta_schedule(timesteps, s=0.008):
    """
    Cosine schedule as proposed in https://openreview.net/forum?id=-NEXDKk8gZ.
    """
    steps = timesteps + 1
    x = np.linspace(0, steps, steps)
    alphas_cumprod = np.cos(((x / steps) + s) / (1 + s) * np.pi * 0.5) ** 2
    alphas_cumprod = alphas_cumprod / alphas_cumprod[0]
    betas = 1 - (alphas_cumprod[1:] / alphas_cumprod[:-1])
    return np.clip(betas, a_min=0, a_max=0.999)


class GaussianDiffusion(nn.Module):
    def __init__(
        self,
        denoise_fn,
        input_size,
        beta_end=0.1,
        diff_steps=100,
        loss_type="l2",
        betas=None,
        beta_schedule="linear",
    ):
        super().__init__()
        self.denoise_fn = denoise_fn
        self.input_size = input_size
        self.__scale = None

        if betas is not None:
            betas = (
                betas.detach().cpu().numpy()
                if isinstance(betas, torch.Tensor)
                else betas
            )
        else:
            if beta_schedule == "linear":
                betas = np.linspace(1e-4, beta_end, diff_steps)
            elif beta_schedule == "quad":
                betas = np.linspace(1e-4**0.5, beta_end**0.5, diff_steps) ** 2
            elif beta_schedule == "const":
                betas = beta_end * np.ones(diff_steps)
            elif beta_schedule == "jsd":  # 1/T, 1/(T-1), 1/(T-2), ..., 1
                betas = 1.0 / np.linspace(diff_steps, 1, diff_steps)
            elif beta_schedule == "sigmoid":
                betas = np.linspace(-6, 6, diff_steps)
                betas = (beta_end - 1e-4) / (np.exp(-betas) + 1) + 1e-4
            elif beta_schedule == "cosine":
                betas = cosine_beta_schedule(diff_steps)
            else:
                raise NotImplementedError(beta_schedule)

        alphas = 1.0 - betas
        alphas_cumprod = np.cumprod(alphas, axis=0)
        alphas_cumprod_prev = np.append(1.0, alphas_cumprod[:-1])

        (timesteps,) = betas.shape
        self.num_timesteps = int(timesteps)
        self.loss_type = loss_type

        to_torch = partial(torch.tensor, dtype=torch.float32)

        self.register_buffer("betas", to_torch(betas))
        self.register_buffer("alphas_cumprod", to_torch(alphas_cumprod))
        self.register_buffer(
            "alphas_cumprod_prev", to_torch(alphas_cumprod_prev)
        )

        # calculations for diffusion q(x_t | x_{t-1}) and others
        self.register_buffer(
            "sqrt_alphas_cumprod", to_torch(np.sqrt(alphas_cumprod))
        )
        self.register_buffer(
            "sqrt_one_minus_alphas_cumprod",
            to_torch(np.sqrt(1.0 - alphas_cumprod)),
        )
        self.register_buffer(
            "log_one_minus_alphas_cumprod",
            to_torch(np.log(1.0 - alphas_cumprod)),
        )
        self.register_buffer(
            "sqrt_recip_alphas_cumprod",
            to_torch(np.sqrt(1.0 / alphas_cumprod)),
        )
        self.register_buffer(
            "sqrt_recipm1_alphas_cumprod",
            to_torch(np.sqrt(1.0 / alphas_cumprod - 1)),
        )

        # calculations for posterior q(x_{t-1} | x_t, x_0)
        posterior_variance = (
            betas * (1.0 - alphas_cumprod_prev) / (1.0 - alphas_cumprod)
        )
        # above: equal to 1. / (1. / (1. - alpha_cumprod_tm1) + alpha_t / beta_t)
        self.register_buffer(
            "posterior_variance", to_torch(posterior_variance)
        )
        # below: log calculation clipped because the posterior variance is 0 at the beginning of the diffusion chain
        self.register_buffer(
            "posterior_log_variance_clipped",
            to_torch(np.log(np.maximum(posterior_variance, 1e-20))),
        )
        self.register_buffer(
            "posterior_mean_coef1",
            to_torch(
                betas * np.sqrt(alphas_cumprod_prev) / (1.0 - alphas_cumprod)
            ),
        )
        self.register_buffer(
            "posterior_mean_coef2",
            to_torch(
                (1.0 - alphas_cumprod_prev)
                * np.sqrt(alphas)
                / (1.0 - alphas_cumprod)
            ),
        )

    @property
    def scale(self):
        return self.__scale

    @scale.setter
    def scale(self, scale):
        self.__scale = scale

    def q_mean_variance(self, x_start, t):
        mean = extract(self.sqrt_alphas_cumprod, t, x_start.shape) * x_start
        variance = extract(1.0 - self.alphas_cumprod, t, x_start.shape)
        log_variance = extract(
            self.log_one_minus_alphas_cumprod, t, x_start.shape
        )
        return mean, variance, log_variance

    def predict_start_from_noise(self, x_t, t, noise):
        return (
            extract(self.sqrt_recip_alphas_cumprod, t, x_t.shape) * x_t
            - extract(self.sqrt_recipm1_alphas_cumprod, t, x_t.shape) * noise
        )

    def q_posterior(self, x_start, x_t, t):
        posterior_mean = (
            extract(self.posterior_mean_coef1, t, x_t.shape) * x_start
            + extract(self.posterior_mean_coef2, t, x_t.shape) * x_t
        )
        posterior_variance = extract(self.posterior_variance, t, x_t.shape)
        posterior_log_variance_clipped = extract(
            self.posterior_log_variance_clipped, t, x_t.shape
        )
        return (
            posterior_mean,
            posterior_variance,
            posterior_log_variance_clipped,
        )

    def p_mean_variance(self, x, cond, t, clip_denoised: bool):
        x_recon = self.predict_start_from_noise(
            x, t=t, noise=self.denoise_fn(x, t, cond=cond)
        )

        if clip_denoised:
            x_recon.clamp_(-1.0, 1.0)

        (
            model_mean,
            posterior_variance,
            posterior_log_variance,
        ) = self.q_posterior(x_start=x_recon, x_t=x, t=t)
        return model_mean, posterior_variance, posterior_log_variance

    @torch.no_grad()
    def p_sample(self, x, cond, t, clip_denoised=False, repeat_noise=False):
        b, *_, device = *x.shape, x.device
        model_mean, _, model_log_variance = self.p_mean_variance(
            x=x, cond=cond, t=t, clip_denoised=clip_denoised
        )
        noise = noise_like(x.shape, device, repeat_noise)
        # no noise when t == 0
        nonzero_mask = (1 - (t == 0).float()).reshape(
            b, *((1,) * (len(x.shape) - 1))
        )
        return (
            model_mean
            + nonzero_mask * (0.5 * model_log_variance).exp() * noise
        )

    @torch.no_grad()
    def p_sample_loop(self, shape, cond):
        device = self.betas.device

        b = shape[0]
        img = torch.randn(shape, device=device)

        for i in reversed(range(0, self.num_timesteps)):
            img = self.p_sample(
                img, cond, torch.full((b,), i, device=device, dtype=torch.long)
            )
        return img

    @torch.no_grad()
    def sample(self, sample_shape=torch.Size(), cond=None):
        if cond is not None:
            shape = cond.shape[:-1] + (self.input_size,)
            # TODO reshape cond to (B*T, 1, -1)
        else:
            shape = sample_shape
        x_hat = self.p_sample_loop(
            shape, cond
        )  # TODO reshape x_hat to (B,T,-1)

        if self.scale is not None:
            x_hat *= self.scale
        return x_hat

    @torch.no_grad()
    def interpolate(self, x1, x2, t=None, lam=0.5):
        b, *_, device = *x1.shape, x1.device
        t = default(t, self.num_timesteps - 1)

        assert x1.shape == x2.shape

        t_batched = torch.stack([torch.tensor(t, device=device)] * b)
        xt1, xt2 = map(lambda x: self.q_sample(x, t=t_batched), (x1, x2))

        img = (1 - lam) * xt1 + lam * xt2
        for i in reversed(range(0, t)):
            img = self.p_sample(
                img, torch.full((b,), i, device=device, dtype=torch.long)
            )

        return img

    def q_sample(self, x_start, t, noise=None):
        noise = default(noise, lambda: torch.randn_like(x_start))

        return (
            extract(self.sqrt_alphas_cumprod, t, x_start.shape) * x_start
            + extract(self.sqrt_one_minus_alphas_cumprod, t, x_start.shape)
            * noise
        )

    def p_losses(self, x_start, cond, t, noise=None):
        noise = default(noise, lambda: torch.randn_like(x_start))

        x_noisy = self.q_sample(x_start=x_start, t=t, noise=noise)
        x_recon = self.denoise_fn(x_noisy, t, cond=cond)

        if self.loss_type == "l1":
            loss = F.l1_loss(x_recon, noise)
        elif self.loss_type == "l2":
            loss = F.mse_loss(x_recon, noise)
        elif self.loss_type == "huber":
            loss = F.smooth_l1_loss(x_recon, noise)
        else:
            raise NotImplementedError()

        return loss

    def log_prob(self, x, cond, *args, **kwargs):
        if self.scale is not None:
            x /= self.scale

        B, T, _ = x.shape

        time = torch.randint(
            0, self.num_timesteps, (B * T,), device=x.device
        ).long()
        loss = self.p_losses(
            x.reshape(B * T, 1, -1),
            cond.reshape(B * T, 1, -1),
            time,
            *args,
            **kwargs,
        )

        return loss



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


import torch as pt
from torch import Tensor, nn
from torch.nn import init, functional as F


class SimpleDiscriminator(nn.Module):
    def __init__(
        self,
        d_model: int,
        n_layer: int,
        n_hidden: int = 1,
    ) -> None:
        super(SimpleDiscriminator, self).__init__()
        self.d_model = d_model
        self.d_hidden = d_model * n_hidden
        self.n_layer = n_layer

        self.weights = nn.ParameterList()
        self.biases = nn.ParameterList()
        for layer in range(n_layer):
            d_input = self.d_model if layer == 0 else self.d_hidden
            d_output = self.d_model if layer == n_layer - 1 else self.d_hidden
            self.weights.append(nn.Parameter(Tensor(d_output, d_input)))
            self.biases.append(nn.Parameter(Tensor(d_output)))
        self.class_weight = nn.Parameter(Tensor(1, self.d_model))
        self.class_bias = nn.Parameter(Tensor(1))

        self._reset_parameters()

    def _reset_parameters(self):
        for layer in range(self.n_layer):
            init.normal_(self.weights[layer], std=1e-1)
            init.zeros_(self.biases[layer])
        init.normal_(self.class_weight, std=1e-1)
        init.zeros_(self.class_bias)

    def forward(self, x: Tensor):
        for layer in range(self.n_layer):
            x = F.linear(x, self.weights[layer], self.biases[layer])
            x = F.gelu(x)
        x = F.linear(x, self.class_weight, self.class_bias)
        x = pt.sigmoid(x)
        x = x.add(1e-10).log()
        return x



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations
from dataclasses import dataclass
from typing import Dict, List
import numpy as np
from tsbench.config.dataset import EvaluationDataset
from tsbench.evaluations.metrics import Metric, Performance
from .metrics import (
    abs_error_sum,
    abs_target_mean,
    abs_target_sum,
    mase,
    naive_error,
    ncrps,
    rmse,
    smape,
)
from .quantile import QuantileForecasts


def evaluate_forecasts(
    forecasts: QuantileForecasts, data: EvaluationDataset
) -> Evaluation:
    """
    Evaluates the forecasts on the provided dataset and returns the metrics
    averaged over all time series.

    Args:
        forecasts: The per time series forecasts. The forecasts *must* align with the time series
            of the given dataset. Otherwise, behavior is undefined.
        dataset: The dataset for which to evaluate the metrics.

    Returns:
        The evaluation of the forecasts.
    """
    assert len(forecasts) > 0, "At least one forecast must be given."
    assert len(forecasts) == len(data.future), (
        "The number of forecasts does not match the number of time series in"
        " the dataset."
    )

    # Compute seasonal error for MASE computation
    seasonal_error = naive_error(data.past, forecasts.seasonality)

    # Compute all the metrics
    results = {
        "nd": abs_error_sum(forecasts.median, data.future)
        / abs_target_sum(data.future),
        "nrmse": rmse(forecasts.median, data.future)
        / abs_target_mean(data.future),
        "mase": mase(forecasts.median, data.future, seasonal_error),
        "smape": smape(forecasts.median, data.future),
        "ncrps": ncrps(forecasts, data.future),
    }

    return Evaluation(results)


# -------------------------------------------------------------------------------------------------


@dataclass
class Evaluation:
    """
    An evaluation instance provides metrics per time series as well as overall
    metrics.
    """

    summary: dict[str, float]
    """
    The metrics summarizing the overall performance of the model.
    """

    @classmethod
    def performance(cls, evaluations: list[Evaluation]) -> Performance:
        """
        Aggregates the provided evaluations into a single performance object.

        Args:
            evaluations: The evaluations.

        Returns:
            The performance object. Since this is not part of the evaluation, it has the
                `num_model_parameters` attribute unset (set to zero).
        """
        metrics = [e.summary for e in evaluations]
        kwargs = {
            m: (
                Metric(0, 0)
                if m == "num_model_parameters"
                else Metric(
                    np.mean(
                        [
                            metric[m] if m in metric else np.nan
                            for metric in metrics
                        ]
                    ),
                    np.std(
                        [
                            metric[m] if m in metric else np.nan
                            for metric in metrics
                        ]
                    ),
                )
            )
            for m in Performance.metrics()
        }
        return Performance(**kwargs)  # type: ignore



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pytorch_lightning.loggers import CSVLogger
from pytorch_lightning import Trainer
import numpy as np
from pathlib import Path


def get_save_dir_from_csvlogger(logger: CSVLogger) -> Path:
    return Path(logger.save_dir) / logger.name / f"version_{logger.version}"


def get_loss_steps(loss_name: str, trainer: Trainer):
    loss = trainer.logger.experiment.metrics
    ep = trainer.current_epoch
    out = [l[loss_name] for l in loss if loss_name in l]
    start = 1 if "val" in loss_name else 0
    return out, np.linspace(start, ep, len(out))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# pylint: disable=missing-function-docstring
import math
import numpy as np
import numpy.ma as ma
from typing import List, Dict

from meta.datasets.splits import EvaluationDataset


def rmse(y_pred: np.ndarray, y_true: np.ndarray) -> float:
    return np.sqrt(((y_pred - y_true) ** 2).mean())


def abs_error_sum(y_pred: np.ndarray, y_true: np.ndarray) -> float:
    return np.abs(y_pred - y_true).sum()


def abs_target_sum(y_true: np.ndarray) -> float:
    return np.sum(np.abs(y_true))


def abs_target_mean(y_true: np.ndarray) -> float:
    return np.mean(np.abs(y_true))


def naive_error(y_past: ma.masked_array, seasonality: int) -> np.ndarray:
    error = np.abs(y_past[:, seasonality:] - y_past[:, :-seasonality]).mean(1)
    return ma.getdata(error)


def mase(
    y_pred: np.ndarray, y_true: np.ndarray, error: np.ndarray
) -> np.ndarray:
    mase_values = np.abs(y_pred - y_true).mean(1) / error
    return mase_values.mean()


def smape(y_pred: np.ndarray, y_true: np.ndarray) -> np.ndarray:
    median = y_pred
    num = np.abs(y_true - median)
    denom = (np.abs(y_true) + np.abs(median)) / 2
    # If the denominator is 0, we set it to float('inf') such that any division yields 0 (this
    # might not be fully mathematically correct, but at least we don't get NaNs)
    denom[denom == 0] = math.inf
    return np.mean(num / denom, axis=1).mean()


def mean_weighted_quantile_loss(
    y_pred: np.ndarray, y_true: np.ndarray, quantiles: List[str]
) -> float:
    y_true_rep = y_true[:, None].repeat(len(quantiles), axis=1)
    quantiles = np.array([float(q) for q in quantiles])
    quantile_losses = 2 * np.sum(
        np.abs(
            (y_pred - y_true_rep)
            * ((y_true_rep <= y_pred) - quantiles[:, None])
        ),
        axis=-1,
    )  # shape [num_time_series, num_quantiles]
    denom = np.sum(np.abs(y_true))  # shape [1]
    weighted_losses = quantile_losses.sum(0) / denom  # shape [num_quantiles]
    return weighted_losses.mean()


def compute_metrics(
    forecasts: np.ndarray,
    dataset: EvaluationDataset,
    quantiles: List[str],
    seasonality: int,
) -> Dict[str, float]:
    """
    Evaluates the forecasts on the provided dataset and returns summary
    metrics.

    Parameters
    ----------
    forecasts: np.ndarray, shape: [n_time_series, prediction_length, n_quantiles]
        The per time-series forecasts. The forecasts *must* align with the time series of the
        given dataset. Otherwise, behavior is undefined.
    dataset: EvaluationDataset
        The dataset for which to evaluate the metrics.
    quantiles: List[str]
        A list containing the quantiles predicted in the forecast
    seasonality: int
        The seasonality of the data.

    Returns
    -------
    Dict
        The evaluation of the forecasts.
    """
    assert len(forecasts) > 0, "At least one forecast must be given."
    assert len(forecasts) == len(
        dataset.future
    ), "The number of forecasts does not match the number of time series in the dataset."
    median = forecasts[:, :, forecasts.shape[-1] // 2]
    # Compute seasonal error for MASE computation
    seasonal_error = naive_error(dataset.past, seasonality)

    # Compute all the metrics
    results = {
        # normalized deviation (see e.g. DeepAR paper)
        "nd": abs_error_sum(median, dataset.future)
        / abs_target_sum(dataset.future),
        "nrmse": rmse(median, dataset.future)
        / abs_target_mean(dataset.future),
        "mase": mase(median, dataset.future, seasonal_error),
        "smape": smape(median, dataset.future),
        "mean_weighted_quantile_loss": mean_weighted_quantile_loss(
            forecasts.transpose((0, 2, 1)),
            dataset.future,
            quantiles=quantiles,
        ),
    }
    return results



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Optional, List, Tuple, Union
import math

import numpy as np
import torch as pt
from torch import Tensor, BoolTensor, nn
from torch.nn import init, functional as F

from tslib.nn.attention import Attention
from tslib.nn.attention.posemb import SinusoidalPositionalEmbedding


class InstanceNorm1d(nn.InstanceNorm1d):
    def forward(self, x: Tensor):
        x = x.transpose(1, 2)
        x = super(InstanceNorm1d, self).forward(x)
        x = x.transpose(1, 2)
        return x


class AttentionKernel(Attention):
    """
    Multi-Head Self-Attention Module.

    A kernel function of queries and keys, which is projection of sequence encodings

    Parameters
    ----------------
    d_hidden: int
        hidden size
    n_head: int
        number of attention heads
    n_enc_layer: int
        number of MLP hidden layers in query/key projections
    symmetric: bool, by default False
        if True, queries and keys are produced by a shared module
    share_values: bool, by default False
        if True, attention heads operate on the same value encodings
    dropout: float, by default 0.0
        dropout rate of normalized attention weights
    temperature: float, by default 1.0
        softmax temperature for attention weight normalization

    Args
    ------------------
    shape: Tensor [N, *, d_hidden]
        sequence encodings for queries and keys
    inter_value: Tensor [N, *, d_hidden] or [N, *, d_head]
        value encodings for interpolation
    extra_value: Tensor [N, *, d_hidden] or [N, *, d_head]
        value encodings for extrapolation
    mask: BoolTensor [N, *]
        mask over input sequence (1 for masked position)
    """

    def __init__(
        self,
        d_hidden: int,
        n_head: int,
        n_enc_layer: int,
        n_dec_layer: int,
        m_output: int = 2,
        symmetric: bool = False,
        share_values: bool = False,
        dropout: float = 0.0,
        temperature: float = 1.0,
        norm: str = "none",
        add_dist: bool = True,
    ) -> None:
        super(AttentionKernel, self).__init__(
            d_hidden=d_hidden,
            n_head=n_head,
            bias=False,
            dropout=dropout,
            temperature=temperature,
        )
        self.d_interm = self.d_hidden * m_output
        self.n_enc_layer = n_enc_layer
        self.n_dec_layer = n_dec_layer
        self.share_values = share_values
        self.symmetric = symmetric
        self.bandwidth = math.sqrt(self.d_head)
        self.add_dist = add_dist

        self._query_weights = nn.ParameterList()
        self._query_biases = nn.ParameterList() if self.bias else []
        self._key_weights = nn.ParameterList()
        self._key_biases = nn.ParameterList() if self.bias else []

        for layer in range(n_enc_layer + 1):
            _query_weight = nn.Parameter(Tensor(d_hidden, d_hidden))
            self._query_weights.append(_query_weight)
            self._key_weights.append(
                (
                    _query_weight
                    if self.symmetric
                    else nn.Parameter(Tensor(d_hidden, d_hidden))
                ),
            )
            if self.bias:
                _query_bias = nn.Parameter(Tensor(d_hidden))
                self._query_biases.append(_query_bias)
                self._key_biases.append(
                    (
                        _query_bias
                        if self.symmetric
                        else nn.Parameter(Tensor(d_hidden))
                    ),
                )
            else:
                self._query_biases.append(None)
                self._key_biases.append(None)

        self._output_weights = nn.ParameterList()
        self._output_biases = nn.ParameterList() if self.bias else []
        for layer in range(n_dec_layer):
            d_input = self.d_hidden if layer == 0 else self.d_interm
            d_output = (
                self.d_hidden if layer == n_dec_layer - 1 else self.d_interm
            )
            self._output_weights.append(
                nn.Parameter(Tensor(d_output, d_input))
            )
            self._output_biases.append(
                nn.Parameter(Tensor(d_output)) if self.bias else None
            )

        if norm == "none":
            # self.enc_norm = nn.Identity()
            self.attn_norm = nn.Identity()
            self.dec_norm = nn.Identity()
        elif norm == "layer":
            # self.enc_norm = nn.LayerNorm(self.d_hidden, elementwise_affine=False)
            self.attn_norm = nn.LayerNorm(
                self.d_hidden, elementwise_affine=False
            )
            self.dec_norm = nn.LayerNorm(
                self.d_hidden, elementwise_affine=False
            )
        elif norm == "instance":
            # self.enc_norm = nn.InstanceNorm1d(self.d_hidden, affine=False)
            self.attn_norm = InstanceNorm1d(self.d_hidden, affine=False)
            self.dec_norm = InstanceNorm1d(self.d_hidden, affine=False)
        else:
            raise ValueError(f"Invalid normalization type: {norm}")

        if self.add_dist:
            self.distance_emb = SinusoidalPositionalEmbedding(d_hidden)
            self.distance_weight = nn.Parameter(Tensor(d_hidden, d_hidden))
            if self.bias:
                self.distance_bias = nn.Parameter(Tensor(d_hidden))
            else:
                self.register_parameter("distance_bias", None)
            self.distance_key = nn.Parameter(Tensor(1, n_head, self.d_head, 1))

        self._reset_parameters()

        self.register_buffer("_query", None, persistent=False)
        self.register_buffer("_key", None, persistent=False)
        self.register_buffer("_inter_score", None, persistent=False)
        self.register_buffer("_extra_score", None, persistent=False)
        self.register_buffer(
            "_inter_dist_comp", pt.zeros(self.d_head), persistent=False
        )
        self.register_buffer(
            "_extra_dist_comp", pt.zeros(self.d_head), persistent=False
        )

    def _reset_parameters(self):
        for layer in range(self.n_enc_layer + 1):
            init.xavier_uniform_(self._query_weights[layer])
            if not self.symmetric:
                init.xavier_uniform_(self._key_weights[layer])
            if self.bias:
                init.zeros_(self._query_biases[layer])
                if not self.symmetric:
                    init.zeros_(self._key_biases[layer])
        for layer in range(self.n_dec_layer):
            init.xavier_normal_(self._output_weights[layer])
            if self.bias:
                init.zeros_(self._output_biases[layer])
        if self.add_dist:
            init.xavier_normal_(self.distance_key)
            init.xavier_normal_(self.distance_weight)
            if self.bias:
                init.zeros_(self.distance_bias)

    def _kernel(
        self,
        q: Tensor,
        k: Tensor,
    ) -> Tensor:
        raise NotImplementedError

    def _compute_attn_score(
        self,
        shape: Tensor,
        left_pad: int,
        right_pad: int,
        mask: Optional[BoolTensor],
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        q = k = shape
        for layer in range(self.n_enc_layer):
            q = F.linear(
                q, self._query_weights[layer], self._query_biases[layer]
            )
            k = F.linear(k, self._key_weights[layer], self._key_biases[layer])
            q = F.relu(q)
            k = F.relu(k)
        q = F.linear(
            q,
            self._query_weights[self.n_enc_layer],
            self._query_biases[self.n_enc_layer],
        )
        k = F.linear(
            k,
            self._key_weights[self.n_enc_layer],
            self._key_biases[self.n_enc_layer],
        )
        self._query = q
        self._key = k
        q = self._split_head(q)
        k = self._split_head(k)
        score = self._kernel(q, k)

        if self.add_dist:
            seq_len = shape.size(1)
            p = pt.arange(
                seq_len, dtype=shape.dtype, device=shape.device
            ).view(1, -1)
            p = self.distance_emb(p)
            p = F.linear(p, self.distance_weight, self.distance_bias)
            p = self._split_head(p)

            idx = pt.arange(seq_len, dtype=pt.long, device=shape.device)
            idx = pt.abs(idx.view(-1, 1) - idx.view(1, -1))
            idx = idx.view(1, 1, seq_len, seq_len).expand_as(score)

            qp = pt.matmul(q, p.transpose(-1, -2))
            qp = pt.gather(qp, dim=-1, index=idx)
            pk = pt.matmul(p, k.transpose(-1, -2))
            pk = pt.gather(pk, dim=-2, index=idx)
            vp = pt.matmul(p, self.distance_key).expand_as(score)
            vp = pt.gather(vp, dim=-2, index=idx)
            score = score + qp + pk + vp

        _inter_mask = pt.eye(
            score.size(-1), dtype=pt.bool, device=score.device
        )
        _inter_score = score.masked_fill(_inter_mask, -math.inf)
        if mask is not None:
            _key_mask = mask.unsqueeze(dim=1).unsqueeze(dim=2)
            _inter_score = _inter_score.masked_fill(_key_mask, -math.inf)
        # remove keys with padding
        _extra_score = score[
            ..., left_pad + 1 : -right_pad, left_pad : -(right_pad + 1)
        ]
        _extra_mask = pt.ones_like(_extra_score[0, 0], dtype=pt.bool).triu(
            diagonal=1
        )
        _extra_score = _extra_score.masked_fill(_extra_mask, -math.inf)
        if mask is not None:
            _key_mask = mask[..., (left_pad + right_pad + 1) :]
            _key_mask = _key_mask.unsqueeze(dim=1).unsqueeze(dim=2)
            _extra_score = _extra_score.masked_fill(_key_mask, -math.inf)
        inter_score = pt.softmax(_inter_score / self.temperature, dim=-1)
        inter_score = pt.masked_fill(inter_score, pt.isnan(inter_score), 0.0)
        self._inter_score = inter_score.detach()
        inter_score = self.dropout(inter_score)
        extra_score = pt.softmax(_extra_score / self.temperature, dim=-1)
        extra_score = pt.masked_fill(extra_score, pt.isnan(extra_score), 0.0)
        self._extra_score = extra_score.detach()
        extra_score = self.dropout(extra_score)

        if self.add_dist:
            p = p.unsqueeze(dim=-2).expand(*score.shape, self.d_head)
            idx = idx.unsqueeze(dim=-1).expand(*score.shape, self.d_head)
            p = pt.gather(p, dim=-2, index=idx)
            inter_dist_comp = pt.sum(inter_score.unsqueeze(dim=-1) * p, dim=-2)
            p = p[
                ..., left_pad + 1 : -right_pad, left_pad : -(right_pad + 1), :
            ]
            extra_dist_comp = pt.sum(extra_score.unsqueeze(dim=-1) * p, dim=-2)
            self._inter_dist_comp = inter_dist_comp.detach()
            self._extra_dist_comp = extra_dist_comp.detach()
        else:
            inter_dist_comp = self._inter_dist_comp
            extra_dist_comp = self._extra_dist_comp

        return inter_score, extra_score, inter_dist_comp, extra_dist_comp

    def _compute_attn_output(
        self,
        score: Tensor,
        value: Tensor,
        dist_comp: Tensor,
    ) -> Tuple[Tensor, Tensor]:
        if self.share_values:
            assert value.size(2) == self.d_head
            value = pt.unsqueeze(value, dim=1)
        else:
            assert value.size(2) == self.d_hidden
            value = self._split_head(value)
        output = score @ value
        output = output + dist_comp
        output = self._merge_head(output)
        output = self.attn_norm(output)
        hidden = output
        for layer in range(self.n_dec_layer):
            hidden = F.linear(
                hidden, self._output_weights[layer], self._output_biases[layer]
            )
            if layer < self.n_dec_layer - 1:
                hidden = F.gelu(hidden)
                hidden = self.dropout(hidden)
            elif layer == self.n_dec_layer - 1:
                output = hidden + output
                output = self.dec_norm(output)
        return output

    def forward(
        self,
        shape: Tensor,
        inter_value: Tensor,
        extra_value: Tensor,
        mask: Optional[BoolTensor],
    ) -> Tuple[Tensor, Tensor]:
        window_size = inter_value.size(1) - extra_value.size(1)
        left_pad = window_size // 2
        right_pad = window_size // 2 - (1 - window_size % 2)
        (
            inter_score,
            extra_score,
            inter_dist_comp,
            extra_dist_comp,
        ) = self._compute_attn_score(shape, left_pad, right_pad, mask)
        interp = self._compute_attn_output(
            inter_score, inter_value, inter_dist_comp
        )
        extrap = self._compute_attn_output(
            extra_score, extra_value, extra_dist_comp
        )
        return interp, extrap


class ExpKernel(AttentionKernel):
    def _kernel(
        self,
        q: Tensor,
        k: Tensor,
    ) -> Tensor:
        return pt.matmul(q, k.transpose(-1, -2)) / self.bandwidth


class RBFKernel(AttentionKernel):
    def _kernel(
        self,
        q: Tensor,
        k: Tensor,
    ) -> Tensor:
        q = q.unsqueeze(dim=-2)
        k = k.unsqueeze(dim=-3)
        score = -pt.sum(pt.pow(q - k, 2), dim=-1) / self.bandwidth
        return score



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Tuple

from .utils import range_overlap


def extend_ranges(
    ranges: List[range], extension_length: int, direction: str = "right"
) -> List[range]:
    """
    Extends a list of ranges by `extension_length` steps.

    Parameters
    ----------
    ranges: List[range]
        Ranges to be extended, non-overlapping and in sorted order.
    extension_length: int
        Positive integer, the length of the extension.
    direction: str
        Direction of the extension. If "right" ("left"), ranges are extended towards
        the right (left). Default "right".
    Returns
    -------
    out_ranges: List[range]
    """
    if not ranges:
        return []
    assert extension_length >= 0, "`extension_length` must be zero or above"
    assert direction in [
        "left",
        "right",
    ], "`direction` must be one of 'left' or 'right'."

    out_ranges = []

    # iterate over real ranges (ground truth anomalies) and "extend" them to include
    # slack windows where it's OK if they're caught
    for i, old_range in enumerate(ranges):
        if direction == "right":
            range_end_ub = (
                ranges[i + 1].start
                if i < len(ranges) - 1
                else ranges[-1].stop + extension_length + 1
            )
            new_range = range(
                old_range.start,
                min(range_end_ub, old_range.stop + extension_length),
            )
        else:
            range_start_lb = ranges[i - 1].stop if i > 0 else 0
            new_range = range(
                max(range_start_lb, old_range.start - extension_length),
                old_range.stop,
            )

        out_ranges.append(new_range)

    return out_ranges


def buffered_precision_recall(
    real_ranges: List[range],
    pred_ranges: List[range],
    buffer_length: int = 5,
) -> Tuple[float, float]:
    """
    Implements a new range-based precision recall metric, that measures how
    well anomalies (`real_ranges`) are caught with labels (`pred_ranges`).

    We extend anomaly ranges by a number of time steps (`buffer_length`) to accomodate
    those raised with a lag. For example, if an annotator has marked `range(5, 9)`, and the
    model has labeled `range(11, 13)` as an anomaly, we would often like to mark this as a
    correctly raised anomaly. There are two reasons for this. (i) Human annotators often draw
    boxes around anomalies with a certain "margin," i.e., with a lead and a lag around the
    true anomaly. (ii) The low-pass filter raises anomalies with a certain latency.

    Therefore, this function looks for intersections between "extended" anomaly ranges,
    those with a buffer of `buffer_length` added after the annotated range, and the predicted
    ranges. Any intersection is counted as a success. More precisely,

    - If an "extended" anomaly range intersects with any labeled range, it's "caught." If an
      anomaly intersects with no predicted range, it's not caught. Recall is,
      `n_caught_anomalies / n_all_anomalies`.
    - If a predicted range intersects with any "extended" anomaly range, it's a good alarm.
      Precision is `n_good_pred_ranges / n_pred_ranges`.

    Note that the numerators (numbers of true positives) are different for precision
    and recall. This is since an anomaly can be caught by multiple pred ranges, as well as a
    pred range marking two separate anomalies. This function allows for this behavior.
    Moreover, a prediction range is either "good" (it intersects with an extended anomaly range,
    and is a "true positive predicted range") or or "bad" (false positive). This is different
    than `segment_precision_recall`, since there a predicted range is counted towards true
    positives and false positives at the same time if it spans the intersection of an
    anomaly segment and a non-anomaly segment.

    Parameters
    ----------
    real_ranges: List[range]
        Python range objects representing ground truth anomalies (e.g., as annotated by
        human labelers). Ranges must ve non-overlapping and in sorted order.
    pred_ranges: List[range]
        Python range objects representing labels produced by the model. Ranges must be
        non-overlapping and in sorted order.
    buffer_length: int
        The number of time periods which a predicted range is allowed lag after an
        anomaly, for which it will be marked as a "good" raise. For example, if the
        actual range is `range(5,7)` and the predicted range is `range(8, 9)`, this prediction
        will be deemed accurate with a buffer length of 2 or above.

    Returns
    -------
    precision: float
        Precision. Ratio of predicted ranges that overlap with an (extended) anomaly range.
    recall: float
        Recall. Ratio of (extended) anomaly ranges that were caught by (overlaps with)
        at least one prediction range.
    """
    if len(real_ranges) == 0 and len(pred_ranges) == 0:
        return 1.0, 1.0
    if len(pred_ranges) == 0:
        return 1.0, 0.0
    if len(real_ranges) == 0:
        return 0.0, 1.0

    extended_ranges = extend_ranges(real_ranges, buffer_length)

    recall_tp, fn = 0, 0
    for tr in extended_ranges:
        # any labels raised this target range?
        raised = any(range_overlap(tr, pr) for pr in pred_ranges)
        recall_tp += raised
        fn += 1 - raised

    precision_tp, fp = 0, 0
    for pr in pred_ranges:
        # any anomalies caught by the label?
        caught = any(range_overlap(tr, pr) for tr in extended_ranges)
        precision_tp += caught
        fp += 1 - caught

    return precision_tp / len(pred_ranges), recall_tp / len(extended_ranges)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Iterable, Iterator
from pathlib import Path
import math

from tqdm import tqdm
from torch import Tensor
from torch import nn
from torch.utils.data import Dataset, DataLoader
import torch as pt
import numpy as np

if TYPE_CHECKING:
    from ..dataset import ForecastingDataset, MetaDataset
    from ..metrics import MeterDict
    from .trainer import Trainer


class Evaluator(object):
    def __init__(
        self,
        dataset: MetaDataset,
        model: nn.Module,
        metrics: MeterDict,
        log_dir: Path,
        batch_size: int,
        n_loader_workers: int = 0,
        cuda_device: int = 0,
        debug: bool = False,
        model_tag: str = "best",
    ):
        self.dataset = dataset
        self.metrics = metrics
        self.log_dir = log_dir
        self.batch_size = batch_size
        self.n_loader_workers = n_loader_workers
        self.cuda_device = cuda_device if pt.cuda.is_available() else -1
        self.pin_memory = self.cuda_device >= 0
        self.debug = debug
        if self.cuda_device >= 0:
            pt.cuda.set_device(self.cuda_device)
            self.model = model.cuda(self.cuda_device)
        else:
            self.model = model.cpu()
        self.load(model_tag)

    def load(self, tag: str):
        try:
            path = self.log_dir.joinpath(f"{tag}.pt.tar")
            state = pt.load(
                path,
                map_location=(
                    f"cuda:{self.cuda_device}"
                    if self.cuda_device >= 0
                    else "cpu"
                ),
            )
            print(f"Load checkpoint from {path}")
        except FileNotFoundError:
            raise ValueError(f"invalid tag {tag}")
        module_state = state.get("model")
        if module_state is None:
            raise KeyError("the checkpoint does not have model data.")
        self.model.load_state_dict(module_state)

    @property
    def test_loader(self) -> Iterator:
        return self.dataset.test_loader(
            batch_size=self.batch_size,
            cuda_device=self.cuda_device,
            n_workers=self.n_loader_workers,
        )

    def evaluate(self) -> None:
        device = "cpu" if self.cuda_device < 0 else f"cuda:{self.cuda_device}"
        prog_bar = tqdm(
            desc=f"({device}) test",
            total=int(math.ceil(self.dataset.test_size / self.batch_size)),
            unit="batch",
        )
        _ = self.model.eval()
        for batch, data in enumerate(self.test_loader):
            with pt.no_grad():
                self._test(*data)
            prog_bar.update(1)
            if self.debug and batch > 0:
                break
        prog_bar.close()

    def _test(self, *data) -> None:
        raise NotImplementedError

    def predict(self, dataset: Optional[Dataset] = None) -> np.ndarray:
        if dataset is None:
            loader = self.test_loader
            data_size = self.dataset.test_size
        else:
            loader = MetaDataset._data_loader(
                dataset=dataset,
                batch_size=self.batch_size,
                shuffle=False,
                cuda_device=self.cuda_device,
                is_training=False,
                n_workers=self.n_loader_workers,
                n_batches=None,
            )
            data_size = len(dataset)
        device = "cpu" if self.cuda_device < 0 else f"cuda:{self.cuda_device}"
        prog_bar = tqdm(
            desc=f"({device}) test",
            total=int(math.ceil(data_size / self.batch_size)),
            unit="batch",
        )
        _ = self.model.eval()
        preds = []
        for batch, data in enumerate(loader):
            with pt.no_grad():
                result = self._predict(*data)
            preds.append(result)
            prog_bar.update(1)
            if self.debug and batch > 0:
                break
        prog_bar.close()
        preds = np.concatenate(preds, axis=0)
        return preds

    def _predict(self, *data) -> np.ndarray:
        raise NotImplementedError

    @classmethod
    def from_trainer(cls, trainer: Trainer, model_tag: str):
        return cls(
            dataset=trainer.dataset,
            model=trainer.get_model_core(),
            metrics=trainer.metrics,
            log_dir=trainer.log_dir,
            batch_size=trainer.eval_batch_size,
            n_loader_workers=trainer.n_loader_workers,
            cuda_device=trainer.cuda_device,
            debug=trainer.debug,
            model_tag=model_tag,
        )

    @classmethod
    def from_dump(
        cls,
        dataset: MetaDataset,
        dump_dir: Union[str, Path],
        cuda_device: int,
        model_tag: str,
    ):
        raise NotImplementedError



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Literal, Dict, Any
from pathlib import Path
from functools import cached_property
import shutil
import tempfile
import pickle
import pandas as pd
from tqdm import tqdm
import catch22
import pytorch_lightning as pl
from src.gluonts.dataset.repository.datasets import materialize_dataset
from src.gluonts.dataset.common import MetaData
from torch.utils.data import DataLoader
from sklearn.neighbors import NearestNeighbors

from meta.data.batch import TripletBatch
from meta.data.sampling import SamplingTripletDataset, SequentialTripletDataset
from meta.datasets.splits import DatasetSplits
from meta.data.dataset import TimeSeries
from meta.datasets.preprocessing import (
    Filter,
    AbsoluteValueFilter,
    ConstantTargetFilter,
    EndOfSeriesCutFilter,
    MinLengthFilter,
    read_transform_write,
    ItemIDTransform,
)
from meta.datasets.registry import register_data_module


@register_data_module
class GluonTSDataModule(pl.LightningDataModule):
    """
    A data module which provides an arbitrary dataset from GlounTS.

    Parameters
    ----------
    dataset_name: str
        The GluonTS name of the dataset.
    context_length_multiple: int
        Context length is prediction length x ``context_length_multiple``.
    support_length_multiple: int
        Support time series length is prediction length x ``support_length_multiple``.
    support_set_size: int
        Number of time series in the support set.
    num_queries: int
        Number of queries used for the same support set.
    data_dir: pathlib.Path
        Path to data.
    prediction_length: int
        Number of predicted time steps.
    standardize: bool
        If each time series is standardized.
    batch_size_train: int
        Batch size during training.
    batch_size_val_test: int
        Batch size during validation and testing.
    num_workers: int
        Number of workers.
    seed: int
        The random seed.
    catch22_train: bool
        If catch22 similarity measure is used for support set selection during training.
    catch22_val_test: bool
        If catch22 similarity measure is used for support set selection during validation and testing.
    cheat: float
        Chance of hiding the ground truth in the support set.
    """

    def __init__(
        self,
        dataset_name: str,
        context_length_multiple: int,
        support_length_multiple: int,
        support_set_size: int,
        num_queries: int = 1,
        data_dir: Path = Path.home() / ".mxnet" / "gluon-ts",
        prediction_length: Optional[int] = None,
        standardize: bool = True,
        batch_size_train: int = 128,
        batch_size_val_test: int = 1000,
        num_workers: int = 0,
        seed: Optional[int] = None,
        catch22_train: bool = False,
        catch22_val_test: bool = False,
        cheat: float = 0.0,
    ):
        super().__init__()
        self.dataset_name = dataset_name
        self.data_dir = data_dir
        self.context_length_multiple = context_length_multiple
        self.support_length_multiple = support_length_multiple
        self._prediction_length = prediction_length
        self.support_set_size = support_set_size
        self.num_queries = num_queries
        self.standardize = standardize
        self.splits: DatasetSplits
        self.batch_size_train = batch_size_train
        self.batch_size_val_test = batch_size_val_test
        self.num_workers = num_workers
        self.seed = seed
        self.catch22_train = catch22_train
        self.catch22_val_test = catch22_val_test
        self.cheat = cheat
        assert (
            not cheat or num_queries == 1
        ), "Cheat sampling only allows num_queries = 1"
        assert not (
            (catch22_val_test or catch22_train) and num_queries > 1
        ), "Catch22 support set selection only works with num_queries equal to one."

    @property
    def context_length(self) -> int:
        return self.context_length_multiple * self.prediction_length

    @property
    def support_length(self) -> int:
        return self.support_length_multiple * self.prediction_length

    @property
    def prediction_length(self) -> int:
        return self._prediction_length or self.meta.prediction_length

    @property
    def root(self) -> Path:
        """
        Returns the directory where all the data pertaining to this dataset is
        stored.
        """
        return self.data_dir / "datasets" / self.dataset_name

    @property
    def meta(self) -> MetaData:
        """
        Returns the dataset's metadata.
        """
        return (
            MetaData.parse_file(self.root / "metadata.json")
            if self.root.exists()
            else None
        )

    def sampling_triplet_dataset(
        self, split: Literal["train", "val"]
    ) -> SamplingTripletDataset:
        if split == "train":
            dataset = self.splits.train().data()
        elif split == "val":
            dataset = self.splits.val().data()
        else:
            raise ValueError(
                "test split should not be used with sampling data set"
            )
        return SamplingTripletDataset(
            dataset=dataset,
            context_length=self.context_length,
            support_length=self.support_length,
            prediction_length=self.prediction_length,
            support_set_size=self.support_set_size,
            num_queries=self.num_queries,
            catch22_nn=self.catch22_nn if self.catch22_train else None,
            cheat=self.cheat,
        )

    def sequential_triplet_dataset(
        self,
        split: Literal["val", "test"],
    ) -> SequentialTripletDataset:
        if split == "val":
            dataset = self.splits.val().data()
            support_dataset = None
        elif split == "test":
            dataset = self.splits.test().data()
            support_dataset = self.splits.val().data()
        else:
            raise ValueError(
                "train split should not be used with sequential data set"
            )
        return SequentialTripletDataset(
            dataset=dataset,
            support_dataset=support_dataset,
            context_length=self.context_length,
            support_length=self.support_length,
            prediction_length=self.prediction_length,
            support_set_size=self.support_set_size,
            num_queries=1,
            seed=self.seed + (split == "test") if self.seed else None,
            catch22_nn=self.catch22_nn if self.catch22_val_test else None,
            cheat=self.cheat,
        )

    def setup(self, stage: Optional[str] = None) -> None:
        # a hack
        if self.dataset_name in [
            "wind_farms_without_missing",
            "kdd_cup_2018_without_missing",
            "pedestrian_counts",
        ]:
            self.catch22_tail = 50000
        else:
            self.catch22_tail = 7000

        self.generate()
        self.splits = DatasetSplits(
            self.meta,
            self.root,
            self.dataset_name,
            self.prediction_length,
            self.standardize,
        )
        self.splits.train().prepare()
        self.splits.val().prepare()
        self.splits.test().prepare()

    @cached_property
    def catch22_nn(self):
        with open(self.root / "catch22" / "nn_100.pkl", "rb") as f:
            return pickle.load(f)

    def train_dataloader(self) -> DataLoader[TripletBatch]:
        return DataLoader(
            self.sampling_triplet_dataset("train"),
            collate_fn=TripletBatch.collate,  # type: ignore
            batch_size=self.batch_size_train,
        )

    def val_dataloader(self) -> DataLoader[TripletBatch]:
        return DataLoader(
            self.sequential_triplet_dataset("val"),
            collate_fn=TripletBatch.collate,  # type: ignore
            batch_size=self.batch_size_val_test,
            num_workers=self.num_workers,
            pin_memory=True,
        )

    def test_dataloader(self) -> DataLoader[TripletBatch]:
        return DataLoader(
            self.sequential_triplet_dataset("test"),
            collate_fn=TripletBatch.collate,  # type: ignore
            batch_size=self.batch_size_val_test,
            num_workers=self.num_workers,
            pin_memory=True,
        )

    def generate(self) -> None:
        if not self.root.exists():
            self.root.mkdir(parents=True)

            # Download data and move to our own managed directory
            with tempfile.TemporaryDirectory() as directory:
                self._materialize(Path(directory))
                source = Path(directory) / self.dataset_name

                # Copy and read metadata
                meta_file = self.root / "metadata.json"
                shutil.copyfile(source / "metadata.json", meta_file)
                meta = MetaData.parse_file(meta_file)

                # Copy the data and apply filters
                # We filter all ts with length smaller than 3 * prediction length in the train / val set.
                # Since for the train split the prediction-length-long tail is cropped
                # we have that ts in train split are at least 2 * prediction length long.
                # This makes it possible to predict a full prediction length
                # and have a context window of size prediction length.
                filters = self._filters(
                    meta.prediction_length, 3 * meta.prediction_length
                )
                read_transform_write(
                    self.root / "train" / "data.json",
                    filters=filters
                    + [
                        EndOfSeriesCutFilter(meta.prediction_length),
                        ItemIDTransform(),
                    ],
                    source=source / "train" / "data.json",
                )
                num_train = count_lines(self.root / "train" / "data.json")

                read_transform_write(
                    self.root / "val" / "data.json",
                    filters=filters + [ItemIDTransform(num_train)],
                    source=source / "train" / "data.json",
                )

                # Although we increase the prediction length for the filters here, this does not
                # exclude any more data! The time series is only longer by the prediction length...
                read_transform_write(
                    self.root / "test" / "data.json",
                    filters=self._filters(
                        2 * meta.prediction_length, 4 * meta.prediction_length
                    )
                    + [
                        ItemIDTransform(num_train),
                    ],
                    source=source / "test" / "data.json",
                )
        num_train = count_lines(self.root / "train" / "data.json")
        num_val = count_lines(self.root / "val" / "data.json")
        num_test = count_lines(self.root / "test" / "data.json")

        assert num_train == num_val and (
            num_test % num_val == 0
        ), "Splits do not match."

        # compute catch22 features
        file = self.root / "catch22" / "features.pkl"
        val_data = (
            DatasetSplits(
                self.meta,
                self.root,
                self.dataset_name,
                self.prediction_length,
                _standardize=False,
            )
            .val()
            .data()
        )
        if not file.exists() and (self.catch22_train or self.catch22_val_test):
            file.parent.mkdir(parents=True, exist_ok=True)

            ts_features = []
            # running this in parallel with more workers seems to throw errors
            for ts in tqdm(val_data, desc="generate catch22 features"):
                # only use 5000 times steps tail to compute features
                start = max(len(ts) - self.catch22_tail, 0)
                ts_features.append(get_features(ts[start:]))

            df = pd.DataFrame(ts_features)
            # remove constant features
            df = df.loc[:, (df != df.iloc[0]).any()]
            df.to_pickle(file)

            # compute catch22 nearest neighbours
        file = self.root / "catch22" / "nn_100.pkl"
        if not file.exists() and (self.catch22_train or self.catch22_val_test):
            df = normalize_features(self.catch22())
            num_ts = len(val_data)
            # for more than 5000 ts use kd_tree algo
            algo = "brute" if num_ts <= 5000 else "kd_tree"
            num_nn = 100
            nbrs = NearestNeighbors(
                n_neighbors=min(num_ts, num_nn + 1), algorithm=algo
            ).fit(df)
            distances, indices = nbrs.kneighbors(df)
            with open(file, "wb") as f:
                # exclude the query ts itself
                pickle.dump(indices[:, 1:], f)

    def catch22(self) -> pd.DataFrame:
        """
        Returns the catch22 features of all time series in the dataset.
        """
        file = self.root / "catch22" / "features.pkl"
        return pd.read_pickle(file)

    @classmethod
    def name(cls) -> str:
        return "dm_gluonts"

    def evaluate_model(self, **kwargs):
        pass

    def _filters(
        self, prediction_length: int, min_length: int
    ) -> List[Filter]:
        return [
            ConstantTargetFilter(
                prediction_length, required_length=self.catch22_tail
            ),
            AbsoluteValueFilter(1e18),
            MinLengthFilter(min_length),
        ]

    def _materialize(self, directory: Path) -> None:
        materialize_dataset(self.dataset_name, directory)


def get_features(ts: TimeSeries) -> Dict[str, Any]:
    features = catch22.catch22_all(ts.values)
    return dict(zip(features["names"], features["values"]))


def normalize_features(df: pd.DataFrame) -> pd.DataFrame:
    return (df - df.min()) / (df.max() - df.min())


def count_lines(filename: Path) -> int:
    return sum(1 for line in open(filename))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from collections import OrderedDict

import numpy as np
import os
import logging
from typing import NamedTuple

import pandas as pd
from pathlib import Path

import tarfile
from src.gluonts.dataset.common import Dataset, load_datasets
from src.gluonts.dataset.repository import (
    default_dataset_path,
)


def extract_dataset(dataset_name: str):
    dataset_folder = default_dataset_path / dataset_name

    if os.path.exists(dataset_folder):
        logging.info(f"found local file in {dataset_folder}, skip extracting")
        return

    compressed_data_path = Path("datasets")
    tf = tarfile.open(compressed_data_path / (dataset_name + ".tar.gz"))
    tf.extractall(default_dataset_path)


def pivot_dataset(dataset):
    ds_list = list(dataset)
    return [
        {
            "item": "0",
            "start": ds_list[0]["start"],
            "target": np.vstack([d["target"] for d in ds_list]),
        }
    ]


class MultivariateDatasetInfo(NamedTuple):
    name: str
    train_ds: Dataset
    test_ds: Dataset
    prediction_length: int
    freq: str
    target_dim: int


def make_dataset(
    values: np.ndarray,
    prediction_length: int,
    start: str = "1700-01-01",
    freq: str = "1H",
):
    target_dim = values.shape[0]

    print(
        f"making dataset with {target_dim} dimension and {values.shape[1]} observations."
    )

    start = pd.Timestamp(start, freq)

    train_ds = [
        {"item": "0", "start": start, "target": values[:, :-prediction_length]}
    ]
    test_ds = [{"item": "0", "start": start, "target": values}]

    return MultivariateDatasetInfo(
        name="custom",
        train_ds=train_ds,
        test_ds=test_ds,
        target_dim=target_dim,
        freq=freq,
        prediction_length=prediction_length,
    )


def make_multivariate_dataset(
    dataset_name: str,
    num_test_dates: int,
    prediction_length: int,
    max_target_dim: int = None,
    dataset_benchmark_name: str = None,
):
    """
    :param dataset_name:
    :param num_test_dates:
    :param prediction_length:
    :param max_target_dim:
    :param dataset_benchmark_name: in case the name is different in the repo and in the benchmark, for instance
    'wiki-rolling' and 'wikipedia'
    :return:
    """

    extract_dataset(dataset_name=dataset_name)

    metadata, train_ds, test_ds = load_datasets(
        metadata=default_dataset_path / dataset_name,
        train=default_dataset_path / dataset_name / "train",
        test=default_dataset_path / dataset_name / "test",
    )
    dim = len(train_ds) if max_target_dim is None else max_target_dim
    from multivariate.datasets.grouper import Grouper

    grouper_train = Grouper(max_target_dim=dim)
    grouper_test = Grouper(
        align_data=False, num_test_dates=num_test_dates, max_target_dim=dim
    )
    return MultivariateDatasetInfo(
        (
            dataset_name
            if dataset_benchmark_name is None
            else dataset_benchmark_name
        ),
        grouper_train(train_ds),
        grouper_test(test_ds),
        prediction_length,
        # metadata.time_granularity,
        metadata.freq,
        dim,
    )


def random_periodic(max_target_dim: int = 16, prediction_length: int = 24):
    """

    :param max_target_dim:
    :param prediction_length:
    :param noise_scale: the scale factor for the noise w.r.t the mean of the input, sigma_noise = input_mean*noise_scale
    :param levels: each input timeseries is drawn from normal distributions with mean=l, for each l in levels
    :return:
    """
    num_periods = 100
    levels = np.random.uniform(low=0, high=100, size=(max_target_dim,))
    levels = np.expand_dims(levels, axis=1) * np.ones(
        (max_target_dim, prediction_length)
    )

    noise_level = 1.0
    seasonal_noise = np.random.uniform(
        low=0, high=noise_level, size=(max_target_dim,)
    )

    seasonal_noise = np.expand_dims(seasonal_noise, axis=1) * np.ones(
        (max_target_dim, prediction_length)
    )
    seed_values = np.random.normal(loc=levels, scale=seasonal_noise * levels)

    values = np.hstack([seed_values for _ in range(num_periods)])

    levels = np.hstack([levels for _ in range(num_periods)])
    noise = np.random.normal(
        loc=0 * levels, scale=levels / 5, size=values.shape
    )

    return make_dataset(values + noise, prediction_length)


def electricity(max_target_dim: int = None):
    return make_multivariate_dataset(
        dataset_name="electricity_nips",
        num_test_dates=7,
        prediction_length=24,
        max_target_dim=max_target_dim,
    )


def solar(max_target_dim: int = None):
    return make_multivariate_dataset(
        dataset_name="solar_nips",
        num_test_dates=7,
        prediction_length=24,
        max_target_dim=max_target_dim,
    )


def traffic(max_target_dim: int = None):
    return make_multivariate_dataset(
        dataset_name="traffic_nips",
        num_test_dates=7,
        prediction_length=24,
        max_target_dim=max_target_dim,
    )


def wiki(max_target_dim: int = None):
    return make_multivariate_dataset(
        dataset_name="wiki-rolling_nips",
        dataset_benchmark_name="wikipedia",
        num_test_dates=5,
        prediction_length=30,
        # we dont use 9K timeseries due to OOM issues
        max_target_dim=2000 if max_target_dim is None else max_target_dim,
    )


def exchange_rate(max_target_dim: int = None):
    return make_multivariate_dataset(
        dataset_name="exchange_rate_nips",
        dataset_benchmark_name="exchange_rate_nips",
        num_test_dates=5,
        prediction_length=30,
        max_target_dim=max_target_dim,
    )


def taxi_30min(max_target_dim: int = None):
    """
    Taxi dataset limited to the most active area, with lower and upper bound:
        lb = [ 40.71, -74.01]
        ub = [ 40.8 , -73.95]
    :param max_target_dim:
    :return:
    """
    return make_multivariate_dataset(
        dataset_name="taxi_30min",
        dataset_benchmark_name="taxi_30min",
        # The dataset corresponds to the taxi dataset used in this reference:
        # https://arxiv.org/abs/1910.03002 but only contains 56 evaluation
        # windows. The last evaluation window was removed because there was an
        # overlap of five time steps in the last and the penultimate
        # evaluation window.
        num_test_dates=56,
        prediction_length=24,
        max_target_dim=max_target_dim,
    )


datasets = OrderedDict(
    [
        ("solar", solar),
        ("exchange_rate", exchange_rate),
        ("electricity", electricity),
        ("traffic", traffic),
        ("wikipedia", wiki),
        ("taxi_30min", taxi_30min),
    ]
)

if __name__ == "__main__":
    extract_dataset("electricity_nips")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pandas as pd
import numpy as np

from src.gluonts.mx import SimpleFeedForwardEstimator, Trainer


def test_incremental_training_smoke_mx():
    estimator = SimpleFeedForwardEstimator(
        prediction_length=6,
        trainer=Trainer(epochs=2),
    )

    dataset = [
        {"start": pd.Period("2022-03-04 00", "1H"), "target": np.ones((100,))},
        {"start": pd.Period("2022-04-05 00", "1H"), "target": np.ones((100,))},
    ]

    predictor = estimator.train(dataset)
    _ = estimator.train_from(predictor, dataset)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Callable, Iterable, List, NamedTuple, Optional, Tuple

import numpy as np
from joblib import Parallel, delayed

from . import buffered_precision_recall
from .utils import labels_to_ranges


class PrecisionRecallAndWeights(NamedTuple):
    precisions: np.ndarray
    recalls: np.ndarray
    precision_weights: np.ndarray
    recall_weights: np.ndarray


def singleton_precision_recall(
    true_labels,
    pred_labels,
) -> Tuple[float, float]:
    """

    Parameters
    ----------
    true_labels
        Binary array of true labels
    pred_labels
        Binary array of predicted labels

    Returns
    -------
    precision: float
    recall: float
    """
    precision = 0.0
    recall = 0

    tp = np.sum(true_labels * pred_labels)
    true_cond_p = np.sum(true_labels)
    pred_cond_p = np.sum(pred_labels)

    if pred_cond_p > 0:
        precision = tp / pred_cond_p
    if true_cond_p > 0:
        recall = tp / true_cond_p

    return precision, recall


def precision_recall_curve_per_ts(
    labels: List[bool],
    scores: List[float],
    thresholds: np.ndarray,
    partial_filter: Optional[Callable] = None,
    singleton_curve: bool = False,
    precision_recall_fn: Callable = buffered_precision_recall,
) -> PrecisionRecallAndWeights:
    true_ranges = labels_to_ranges(labels)
    precisions = np.zeros(len(thresholds))
    recalls = np.zeros(len(thresholds))

    precision_weights, recall_weights = (
        np.zeros(len(thresholds)),
        np.zeros(len(thresholds)),
    )

    for ix, th in enumerate(thresholds):
        if partial_filter is None:
            pred_labels = scores >= th
        else:
            pred_labels = partial_filter(th)

        if singleton_curve:
            true_labels_np = np.array(labels, dtype=float)
            pred_labels_np = np.array(pred_labels, dtype=float)
            _prec, _reca = singleton_precision_recall(
                true_labels_np, pred_labels_np
            )
            _prec_w, _reca_w = np.sum(pred_labels_np), np.sum(true_labels_np)
        else:
            pred_ranges = labels_to_ranges(pred_labels)
            _prec, _reca = precision_recall_fn(true_ranges, pred_ranges)
            _prec_w, _reca_w = len(pred_ranges), len(true_ranges)

        precisions[ix] += _prec * _prec_w
        recalls[ix] += _reca * _reca_w

        precision_weights[ix] += _prec_w
        recall_weights[ix] += _reca_w

    return PrecisionRecallAndWeights(
        precisions, recalls, precision_weights, recall_weights
    )


def aggregate_precision_recall_curve(
    label_score_iterable: Iterable,
    thresholds: Optional[np.array] = None,
    partial_filter: Optional[Callable] = None,
    singleton_curve: bool = False,
    precision_recall_fn: Callable = buffered_precision_recall,
    n_jobs: int = -1,
):
    """
    Computes aggregate range-based precision recall curves over a data set,
    iterating over individual time series. Optionally takes partially
    constructed filter that converts given scores/thresholds to anomaly labels.
    See `src.gluonts.nursery.anomaly_detection.supervised_metrics.filters` for
    example filters.

    Parameters
    ----------
    label_score_iterable: Iterable
        An iterable that gives 2-tuples of np.arrays (of identical length),
        corresponding to `true_labels` and `pred_scores` respectively.
    thresholds: np.ndarray
        An np.array of score thresholds for which to compute precision recall values.
        If the filter_type argument is provided, these are the threshold values of
        the filter. If not, they will be applied as a single step hard threshold to
        predicted scores.
    partial_filter: Callable
        Partial constructor for a "filter" object. If provided, this function can be called with a "score_threshold" to
        return labels used for precision and recall computation. If not provided, labels will be assigned with a hard
        threshold.
        See `src.gluonts.nursery.anomaly_detection.supervised_metrics.filters` for example filters.
    singleton_curve: bool
        If true, range-based precision recall will not be computed
    precision_recall_fn:
        Function to call in order to get the precision, recall metrics.
    n_jobs: int
        How many concurrent threads for parallelization, default is -1 (use all cpu available)

    Returns
    -------
    (Same as output of `sklearn.metrics.precision_recall_curve`)
    precision : array, shape = [n_thresholds + 1]
        Precision values such that element i is the precision of
        predictions with score >= thresholds[i] and the last element is 1.

    recall : array, shape = [n_thresholds + 1]
        Decreasing recall values such that element i is the recall of
        predictions with score >= thresholds[i] and the last element is 0.

    thresholds : array, shape = [n_thresholds <= len(np.unique(scores))]
        Increasing thresholds on the decision function used to compute
        precision and recall.
    """
    if thresholds is None:
        thresholds = np.unique(
            np.concatenate([scores for _, scores in label_score_iterable])
        )

    all_metrics = Parallel(n_jobs=n_jobs, verbose=10)(
        delayed(precision_recall_curve_per_ts)(
            labels,
            scores,
            thresholds,
            partial_filter,
            singleton_curve,
            precision_recall_fn,
        )
        for labels, scores in label_score_iterable
    )
    (
        all_precisions,
        all_recalls,
        all_precision_weights,
        all_recall_weights,
    ) = zip(*all_metrics)

    precisions = np.sum(all_precisions, axis=0)
    recalls = np.sum(all_recalls, axis=0)
    precision_weights = np.sum(all_precision_weights, axis=0)
    recall_weights = np.sum(all_recall_weights, axis=0)

    # normalize
    with np.errstate(divide="ignore", invalid="ignore"):
        precisions = np.where(
            precision_weights > 0, precisions / precision_weights, 0.0
        )
        recalls = np.where(recall_weights > 0, recalls / recall_weights, 0.0)

    # Start from the latest threshold where the full recall is attained.
    perfect_recall_ixs = np.where(recalls == 1.0)[0]
    first_ind = perfect_recall_ixs[-1] if len(perfect_recall_ixs) > 0 else 0
    return (
        np.r_[precisions[first_ind:], 1],
        np.r_[recalls[first_ind:], 0],
        thresholds[first_ind:],
    )


def aggregate_precision_recall(
    labels_pred_iterable: Iterable,
    precision_recall_fn: Callable = buffered_precision_recall,
) -> Tuple[float, float]:
    """
    Computes aggregate range-based precision recall metrics for the given
    prediction labels.

    Parameters
    ----------
    labels_pred_iterable
        An iterable that gives 2-tuples of boolean lists corresponding to `true_labels` and
        `pred_labels` respectively.
    precision_recall_fn
        Function to call in order to get the precision, recall metrics.

    Returns
    -------
    A tuple containing average precision and recall in that order.
    """
    total_prec, total_reca, total_prec_w, total_reca_w = 0.0, 0.0, 0.0, 0.0
    for true_labels, pred_labels in labels_pred_iterable:
        true_ranges = labels_to_ranges(true_labels)
        pred_ranges = labels_to_ranges(pred_labels)

        _prec, _reca = precision_recall_fn(true_ranges, pred_ranges)
        _prec_w, _reca_w = len(pred_ranges), len(true_ranges)
        total_prec += _prec * _prec_w
        total_prec_w += _prec_w

        total_reca += _reca * _reca_w
        total_reca_w += _reca_w

    return (
        total_prec / total_prec_w if total_prec_w > 0 else 0,
        total_reca / total_reca_w if total_reca_w > 0 else 0,
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
from pathlib import Path
from typing import Any, Dict, Optional
import catch22
import click
import pandas as pd
from tqdm.auto import tqdm
from tqdm.contrib.concurrent import process_map
from tsbench.config import DATASET_REGISTRY
from tsbench.constants import DEFAULT_DATA_CATCH22_PATH, DEFAULT_DATA_PATH
from ._main import datasets


@datasets.command(short_help="Compute catch22 features.")
@click.option(
    "--dataset",
    type=str,
    default=None,
    help=(
        "The dataset to compute the catch22 features for. "
        "If not provided, computes features for all datasets."
    ),
)
@click.option(
    "--data_path",
    type=click.Path(exists=True),
    default=DEFAULT_DATA_PATH,
    show_default=True,
    help="The local path where datasets are stored.",
)
@click.option(
    "--output_path",
    type=click.Path(),
    default=DEFAULT_DATA_CATCH22_PATH,
    show_default=True,
    help=(
        "The path where catch22 features are written to. "
        "Features are written as `<dataset>.parquet` files."
    ),
)
def compute_catch22(dataset: Optional[str], data_path: str, output_path: str):
    """
    Computes the catch22 features for each time series in a dataset.

    Computations are either run for a single dataset or all datasets in the
    registry.
    """
    target = Path(data_path)
    target.mkdir(parents=True, exist_ok=True)

    if dataset is None:
        dataset_names = [(k, v(target)) for k, v in DATASET_REGISTRY.items()]
    else:
        dataset_names = [(dataset, DATASET_REGISTRY[dataset](target))]

    directory = Path(output_path)
    directory.mkdir(parents=True, exist_ok=True)

    for dataset_name, config in tqdm(
        dataset_names, disable=len(dataset_names) == 1
    ):
        file = directory / f"{dataset_name}.parquet"
        if file.exists():
            continue

        ts_features = process_map(
            _get_features,
            config.data.train(
                val=False
            ).gluonts(),  # Get features on train set
            max_workers=os.cpu_count(),
            desc=dataset_name,
            chunksize=1,
        )
        df = pd.DataFrame(ts_features)
        df.to_parquet(file)  # type: ignore


def _get_features(ts: Dict[str, Any]) -> Dict[str, Any]:
    features = catch22.catch22_all(ts["target"])
    return dict(zip(features["names"], features["values"]))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Dict, Tuple, List, Union
from collections import OrderedDict
from pathlib import Path
from itertools import product
from functools import reduce
import traceback
import textwrap
import socket
import json

import torch as pt
import numpy as np
import pandas as pd

from .distributed import is_main_process, synchronize

if TYPE_CHECKING:
    from ..dataset import MetaDataset


class HyperOptManager(object):
    def __init__(
        self,
        dataset: MetaDataset,
        work_dir: Path,
        fixed_params: Dict,
        cuda_device: int,
        key_factor: str = "loss",
        min_mode: bool = True,
        resume: bool = False,
    ):
        self.dataset = dataset
        self.param_names, self.varied_params = zip(
            *sorted(dataset.hyperparam_space.items(), key=lambda x: x[0])
        )
        self.fixed_params = fixed_params
        self.work_dir = work_dir
        self.work_dir.mkdir(parents=True, exist_ok=True)
        self.cuda_device = cuda_device
        self.key_factor = key_factor
        self.min_mode = min_mode

        self.exp_ids = OrderedDict()
        self.records = OrderedDict()
        if resume:
            self.load_records()

    def __getitem__(self, key: Union[Tuple, int]) -> Union[int, Tuple]:
        if isinstance(key, tuple):
            index = 0
            for i, param in enumerate(key):
                idx = self.varied_params[i].index(param)
                index = index * len(self.varied_params[i]) + idx
            return index
        else:
            params = []
            for vp in reversed(self.varied_params):
                idx = key % len(vp)
                params.insert(0, vp[idx])
                key = key // len(vp)
            return tuple(params)

    def __len__(self):
        return reduce(lambda p, vp: p * len(vp), self.varied_params, 1)

    def hyperparameters(self, n_iter: int):
        n_choices_per_param = [len(vp) for vp in self.varied_params]
        n_choices = reduce(lambda p, x: p * x, n_choices_per_param, 1)
        n_covered = len(self.records)
        if n_iter > n_choices - n_covered:
            raise ValueError(
                f"# iterations {n_iter} exceeds the remaining grids {n_choices-n_covered}"
            )
        samples = list(np.random.choice(n_choices, n_iter + n_covered))
        for params in self.records:
            index = self[params]
            if index in samples:
                samples.remove(index)
        samples = samples[:n_iter]
        for index in samples:
            yield self[index]

    def get_log_dir(self, params) -> str:
        exp_id = f"{self[params]:09d}-{socket.gethostname()}"
        self.exp_ids[params] = exp_id
        log_dir = self.work_dir.joinpath(exp_id)
        log_dir.mkdir(parents=True, exist_ok=True)
        return log_dir

    def get_optimal_model(self) -> Path:
        if len(self.records) == 0:
            raise ValueError("No model has been trained!")
        selector = np.nanargmin if self.min_mode else np.nanargmax
        models = list(self.records.keys())
        values = [r[self.key_factor] for r in self.records.values()]
        dirname = self.exp_ids[models[selector(values)]]
        return self.work_dir.joinpath(dirname)

    def dump_records(self):
        columns = list(self.param_names)
        index = []
        data = []
        for params, record in self.records.items():
            values = list(params)
            for name, value in record.items():
                if name not in columns:
                    columns.append(name)
                values.append(value)
            data.append(values)
            index.append(self.exp_ids[params])
        df = pd.DataFrame(data, index=index, columns=columns)
        df.to_csv(self.work_dir.joinpath("records.csv"))

    def load_records(self):
        print(f"Loading records from {self.work_dir}")
        record_path = self.work_dir.joinpath("records.csv")
        if record_path.exists():
            df = pd.read_csv(record_path, index_col=0)
            for exp_id, record in df.iterrows():
                params = (
                    record.loc[list(self.param_names)]
                    .values.flatten()
                    .tolist()
                )
                params = tuple(
                    eval(x) if isinstance(x, str) else x for x in params
                )
                self.exp_ids[params] = exp_id
                self.records[params] = record.iloc[
                    len(self.param_names) :
                ].to_dict()
            return
        raise RuntimeError("Cannot load previous checkpoint in random search.")

    def update_record(self, params: Tuple, **record):
        if self.key_factor not in record:
            raise ValueError(f"key factor {self.key_factor} is not provided.")
        self.records[params] = record
        self.dump_records()

    @staticmethod
    def print_params(params: Dict):
        print("Current Hyperparams:")
        for k, v in params.items():
            print(f"\t{k:>16}{str(v):>16}")

    def run_training(self, params: Dict) -> Dict:
        raise NotImplementedError

    def run_test(self, exp_dir: Path):
        raise NotImplementedError

    def random_search(self, n_iter: int):
        for params in self.hyperparameters(n_iter):
            configs = dict(zip(self.param_names, params))
            self.print_params(configs)
            configs.update(self.fixed_params)
            log_dir = self.get_log_dir(params)
            if is_main_process():
                with log_dir.joinpath("configs.json").open("w") as f:
                    json.dump(configs, f, indent=4)
            configs["cuda_device"] = self.cuda_device
            configs["log_dir"] = log_dir
            try:
                exp_info = self.run_training(configs)
            except Exception as e:
                print(
                    f"Current config has problem:\n{textwrap.indent(traceback.format_exc(), ' '*4)}skipped."
                )
                continue
            self.update_record(params, **exp_info)
        if is_main_process():
            exp_dir = self.get_optimal_model()
            test_info = self.run_test(exp_dir)
        else:
            test_info = None
        synchronize()
        return test_info

    def fixed_train(self):
        exp_id = f"fixed-{socket.gethostname()}"
        log_dir = self.work_dir.joinpath(exp_id)
        log_dir.mkdir(parents=True, exist_ok=True)
        configs = self.dataset.fixed_hyperparams
        configs.update(self.fixed_params)

        if is_main_process():
            with log_dir.joinpath("configs.json").open("w") as f:
                json.dump(configs, f, indent=4)
        configs["cuda_device"] = self.cuda_device
        configs["log_dir"] = log_dir
        try:
            _ = self.run_training(configs)
        except Exception as e:
            print(
                f"Current config has problem:\n{textwrap.indent(traceback.format_exc(), ' '*4)}skipped."
            )
            return
        test_info = self.run_test(log_dir)
        return test_info



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Any, Dict, Optional, List
import pytorch_lightning as pl
import torch
import torch.optim as optim
from torch import nn
from torchmetrics import Metric
from torch.optim.lr_scheduler import ReduceLROnPlateau

from meta.data.batch import TripletBatch
from meta.models.model import MetaModel
from meta.common.torch import get_mask


class MetaLightningModule(pl.LightningModule):
    """
    PyTorch Lightning module for the meta models.
    """

    def __init__(
        self,
        model: MetaModel,
        val_dataset_names: List[str],
        test_dataset_names: List[str],
        lr_scheduler_monitor: str,
        loss: Metric,
        crps: Metric,
        crps_scaled: Optional[Metric] = None,
        lr: float = 1e-3,
        lr_on_plateau_patience: int = 5,
        lr_on_plateau_factor: float = 0.3,
        quantile_width: Metric = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.model = model
        self.train_loss = loss.clone()
        self.lr = lr
        self.lr_on_plateau_patience = lr_on_plateau_patience
        self.lr_on_plateau_factor = lr_on_plateau_factor
        self.lr_scheduler_monitor = lr_scheduler_monitor
        self.quantile_width = quantile_width

        # validation
        self.val_loss = nn.ModuleList(
            [loss.clone() for _ in range(len(val_dataset_names))]
        )
        self.val_loss_macro = sum(self.val_loss) / len(self.val_loss)

        if quantile_width is not None:
            self.val_quantile_width = nn.ModuleList(
                [quantile_width.clone() for _ in range(len(val_dataset_names))]
            )
            self.test_quantile_width = nn.ModuleList(
                [
                    quantile_width.clone()
                    for _ in range(len(test_dataset_names))
                ]
            )

        self.val_dataset_names = val_dataset_names
        (
            self.val_crps,
            self.val_crps_macro,
            self.val_crps_scaled,
            self.val_crps_scaled_macro,
        ) = _init_metrics(val_dataset_names, crps, crps_scaled)

        # test
        self.test_dataset_names = test_dataset_names
        (
            self.test_crps,
            self.test_crps_macro,
            self.test_crps_scaled,
            self.test_crps_scaled_macro,
        ) = _init_metrics(test_dataset_names, crps, crps_scaled)

    def configure_optimizers(self) -> Dict:
        """
        Example of how to use StepLR:
            {
                "scheduler": StepLR(optimizer, gamma=0.99, ...)
                interval": "epoch",
                "frequency": 1,
            }
        """
        optimizer = optim.Adam(self.parameters(), lr=self.lr)

        return {
            "optimizer": optimizer,
            "lr_scheduler": {
                "scheduler": ReduceLROnPlateau(
                    optimizer,
                    mode="min",
                    patience=self.lr_on_plateau_patience,
                    factor=self.lr_on_plateau_factor,
                    verbose=True,
                ),
                "monitor": self.lr_scheduler_monitor,
                "frequency": 1,
                # If "monitor" references validation metrics, then "frequency" should be set to a
                # multiple of "trainer.check_val_every_n_epoch".
            },
        }

    def training_step(self, batch: TripletBatch, *args: Any) -> torch.Tensor:
        y_true = batch.query_future.sequences
        mask = get_mask(y_true.size()[1], batch.query_future.lengths)
        y_pred = self.model(batch.support_set, batch.query_past)
        loss = self.train_loss(y_pred, y_true, mask)
        self.log("train_loss", self.train_loss)
        return loss

    def validation_step(
        self, batch: TripletBatch, batch_idx, *args: Any
    ) -> None:
        # a hack since pl considers a list of only one dataloader not as multiple
        if len(self.val_dataset_names) == 1:
            dataset_idx = 0
        else:
            dataset_idx = args[0]
        y_true = batch.query_future.sequences
        mask = get_mask(y_true.size()[1], batch.query_future.lengths)
        y_pred = self.model(batch.support_set, batch.query_past)
        d_name = self.val_dataset_names[dataset_idx]
        self.val_loss[dataset_idx](y_pred, batch.query_future.sequences, mask)
        self.log(
            f"{d_name}_val_loss",
            self.val_loss[dataset_idx],
            add_dataloader_idx=False,
        )
        self.val_crps[dataset_idx](
            y_pred,
            batch.query_future.sequences,
            mask,
            batch.query_future.scales,
        )
        self.log(
            f"{d_name}_val_crps",
            self.val_crps[dataset_idx],
            add_dataloader_idx=False,
        )
        if self.val_crps_scaled is not None:
            self.val_crps_scaled[dataset_idx](
                y_pred, batch.query_future.sequences, mask
            )
            self.log(
                f"{d_name}_val_crps_scaled",
                self.val_crps_scaled[dataset_idx],
                add_dataloader_idx=False,
            )

        if self.quantile_width is not None:
            self.val_quantile_width[dataset_idx](
                y_pred, batch.query_future.sequences, mask
            )
            self.log(
                f"{d_name}_val_quantile_width",
                self.val_quantile_width[dataset_idx],
                add_dataloader_idx=False,
            )

    def validation_epoch_end(self, validation_step_outputs):
        self.log("val_crps_macro", self.val_crps_macro, prog_bar=True)
        self.log("val_loss_macro", self.val_loss_macro, prog_bar=True)
        if self.val_crps_scaled:
            self.log(
                "val_crps_scaled_macro",
                self.val_crps_scaled_macro,
                prog_bar=True,
            )

    def test_step(self, batch: TripletBatch, batch_idx, *args: Any) -> None:
        # a hack since pl considers a list of only one dataloader not as multiple
        if len(self.test_dataset_names) == 1:
            dataset_idx = 0
        else:
            dataset_idx = args[0]
        y_true = batch.query_future.sequences
        mask = get_mask(y_true.size()[1], batch.query_future.lengths)
        y_pred = self.model(batch.support_set, batch.query_past)
        d_name = self.test_dataset_names[dataset_idx]
        self.test_crps[dataset_idx](
            y_pred,
            batch.query_future.sequences,
            mask,
            batch.query_future.scales,
        )
        self.log(
            f"{d_name}_test_crps",
            self.test_crps[dataset_idx],
            add_dataloader_idx=False,
        )
        if self.test_crps_scaled is not None:
            self.test_crps_scaled[dataset_idx](
                y_pred, batch.query_future.sequences, mask
            )
            self.log(
                f"{d_name}_test_crps_scaled",
                self.test_crps_scaled[dataset_idx],
                add_dataloader_idx=False,
            )
        if self.quantile_width is not None:
            self.test_quantile_width[dataset_idx](
                y_pred, batch.query_future.sequences, mask
            )
            self.log(
                f"{d_name}_test_quantile_width",
                self.test_quantile_width[dataset_idx],
                add_dataloader_idx=False,
            )

    def test_epoch_end(self, test_step_outputs):
        self.log("test_crps_macro", self.test_crps_macro, prog_bar=True)
        if self.test_crps_scaled:
            self.log(
                "test_crps_scaled_macro",
                self.test_crps_scaled_macro,
                prog_bar=True,
            )


def _init_metrics(
    dataset_names: List[str],
    crps: Metric,
    crps_scaled: Optional[Metric] = None,
):
    n_datasets = len(dataset_names)
    crps = nn.ModuleList([crps.clone() for _ in range(n_datasets)])
    macro_crps = sum(crps) / len(crps)
    if crps_scaled:
        crps_scaled = nn.ModuleList(
            [crps_scaled.clone() for _ in range(n_datasets)]
        )
        macro_crps_scaled = sum(crps_scaled) / len(crps_scaled)
    return (
        crps,
        macro_crps,
        crps_scaled,
        macro_crps_scaled if crps_scaled else None,
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from __future__ import annotations
from typing import List, Tuple, Dict, Optional

import numpy as np
from tqdm import tqdm

from .timeseries import TimeSeries, TimeSeriesCorpus
from .windows import WindowsDataset


class ForecastingDataset(WindowsDataset):
    def __init__(
        self,
        corpus: TimeSeriesCorpus,
        windows: List[Tuple[int, int]],
        horizon: int,
        context: Optional[int] = None,
        gap: int = 0,
    ) -> None:
        super(ForecastingDataset, self).__init__(
            corpus=corpus,
            windows=windows,
        )
        self.horizon = horizon
        self.context = context
        self.gap = gap

    def __getitem__(self, index: int) -> Tuple[TimeSeries, TimeSeries]:
        ts, ctx_window, tgt_window = self._fetch(index)
        ctx = ts[ctx_window]
        tgt = ts[tgt_window]
        return ctx, tgt

    def _fetch(self, index: int) -> Tuple[TimeSeries, slice, slice]:
        scope_id, tgt_head = self.windows[index]
        tgt_tail = tgt_head + self.horizon
        ctx_tail = tgt_head - self.gap
        ctx_head = None if self.context is None else ctx_tail - self.context
        ctx_window = slice(ctx_head, ctx_tail)
        tgt_window = slice(tgt_head, tgt_tail)
        return self.corpus[scope_id], ctx_window, tgt_window

    @classmethod
    def sliding_windows(
        cls,
        corpus: TimeSeriesCorpus,
        horizon: int,
        context: Optional[int] = None,
        gap: int = 0,
        shift: Optional[int] = None,
    ):
        shift = shift or horizon
        windows = []
        for scope_id, series in tqdm(
            enumerate(corpus),
            desc="Building dataset",
            total=len(corpus),
        ):
            stop = len(series) - horizon
            start = gap
            if context is not None:
                start += context
            windows.extend(
                [(scope_id, index) for index in np.arange(stop, start, -shift)]
            )
        return cls(
            corpus,
            windows=windows,
            horizon=horizon,
            context=context,
            gap=gap,
        )

    @classmethod
    def random_windows(
        cls,
        corpus: TimeSeriesCorpus,
        n_windows: int,
        horizon: int,
        context: Optional[int] = None,
        gap: int = 0,
        shift: int = 1,
        cold_start: bool = False,
        favor_recent: bool = False,
    ):
        all_windows = []
        weights = []
        for scope_id, series in tqdm(
            enumerate(corpus),
            desc="Building dataset",
            total=len(corpus),
        ):
            stop = len(series) - horizon
            start = gap
            if context is not None and not cold_start:
                start = start + context
            else:
                start = start + 1
            windows = [
                (scope_id, index) for index in range(stop, start, -shift)
            ]
            all_windows.extend(windows)
            if favor_recent:
                weights.extend(range(stop, start, -shift))
            else:
                weights.extend([1.0] * len(windows))
        weights = np.array(weights) / np.sum(weights)
        np.random.seed(42)
        windows_idx = np.random.choice(
            len(all_windows),
            size=n_windows,
            replace=len(all_windows) < n_windows,
            p=weights,
        )
        windows = [all_windows[i] for i in windows_idx]
        return cls(
            corpus,
            windows=windows,
            horizon=horizon,
            context=context,
            gap=gap,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
from pathlib import Path
import itertools
import yaml
from tqdm.auto import tqdm
import matplotlib.pyplot as plt
import numpy as np
import click
import torch

from meta.datasets import get_data_module, DATASETS_FILTERED
from meta.datasets.cheat import CheatMetaData
from meta.common.torch import tensor_to_np
import pandas as pd


@click.group()
def main():
    return


@main.command()
@click.option(
    "--catch22",
    default=False,
    help="Compute catch22 features and nearest neighbors. Can be time consuming!",
)
def real(catch22: bool):
    """
    Download and process real-world datasets specified in 'DATASETS_FILTERED'.

    Parameters
    ----------
    catch22: bool
        If true, catch22 features for each time series and 100 nearest neighbors w.r.t. l2 distance in
        feature space are computed.
    """
    for dataset in tqdm(DATASETS_FILTERED):
        dm = get_data_module(
            name="dm_" + dataset,
            dataset_name=dataset,
            context_length_multiple=4,
            support_length_multiple=4,
            support_set_size=1,
            catch22_train=catch22,  # compute catch22 features and nearest neighbors
        )
        dm.setup()


@main.command()
@click.argument("config", type=click.Path(exists=True), nargs=1)
def artificial(
    config: str,
):
    """
    Generate artificial datasets specified in a config file.

    Parameters
    ----------
    config: str
        Path to config file.
    """
    with Path(config).open() as f:
        options = yaml.safe_load(f)

    assert len(options["dm_name"]) == 1, len(options["dm_name"])
    dm_name = options["dm_name"][0]
    prefix = f"{dm_name}_"

    data_configs = {}
    for key in options:
        if key in CheatMetaData.__annotations__.keys():
            data_configs[key] = options[key]
        elif key.startswith(prefix):
            data_configs[key[len(prefix) :]] = options[key]

    values = list(itertools.product(*data_configs.values()))
    all_configurations = [dict(zip(data_configs.keys(), v)) for v in values]

    for config in all_configurations:
        dm = get_data_module(
            name=dm_name,
            **config,
        )
        dm.setup()


@main.command()
def statistics():
    """
    Compute statistics of real-world datasets.
    """

    path_to_data = Path.home() / ".mxnet" / "gluon-ts" / "datasets"
    path_save_plots = Path.home() / "data" / "plots"

    datasets = DATASETS_FILTERED

    dataset_stats = {}
    for dataset in tqdm(datasets):
        dm = get_data_module(
            name="dm_" + dataset,
            dataset_name=dataset,
            context_length_multiple=4,
            support_length_multiple=4,
            support_set_size=1,
        )
        dm.setup()
        train_file_size = os.path.getsize(
            path_to_data / dataset / "train" / "data.json"
        )

        dataset_stats[dataset] = {
            "length": len(dm.splits.train().data()),
            "n_total_observations": dm.splits.train()
            .data()
            .number_of_time_steps,
            "freq": dm.meta.freq,
            "train_file_size": train_file_size,
            "prediction_length": dm.meta.prediction_length,
        }

    lengths = np.asarray([dataset_stats[d]["length"] for d in datasets])
    lengths_normalized = lengths / sum(lengths)

    n_total = np.asarray(
        [dataset_stats[d]["n_total_observations"] for d in datasets]
    )
    n_total_normalized = n_total / sum(n_total)

    datasets, lengths_normalized, n_total_normalized = zip(
        *(
            sorted(
                list(zip(datasets, lengths_normalized, n_total_normalized)),
                key=lambda x: x[1],
                reverse=True,
            )
        )
    )

    x = np.arange(len(datasets))

    # -------- plot without logarithmic y-axis ------
    fig, ax = plt.subplots(figsize=[25, 8])
    width = 0.35  # the width of the bars
    rects1 = ax.bar(
        x - width / 2, lengths_normalized, width, label="number of ts"
    )
    rects2 = ax.bar(
        x + width / 2,
        n_total_normalized,
        width,
        label="total number of time steps",
    )

    # Add some text for labels, title and custom x-axis tick labels, etc.
    ax.set_ylabel("normalized size")
    ax.set_title("Datasets")
    ax.set_xticks(x)
    ax.set_xticklabels(datasets, rotation=90)
    ax.legend()
    fig.tight_layout()
    fig.savefig(
        path_save_plots / "datasets.png",
        bbox_inches="tight",
    )
    plt.show()
    plt.close(fig)

    # -------- plot with logarithmic y-axis ------

    fig, ax = plt.subplots(figsize=[25, 8])
    width = 0.35  # the width of the bars
    rects1 = ax.bar(
        x - width / 2, lengths_normalized, width, label="number of ts"
    )
    rects2 = ax.bar(
        x + width / 2,
        n_total_normalized,
        width,
        label="total number of time steps",
    )
    ax.set_yscale("log")
    # Add some text for labels, title and custom x-axis tick labels, etc.
    ax.set_ylabel("normalized size")
    ax.set_title("Datasets")
    ax.set_xticks(x)
    ax.set_xticklabels(datasets, rotation=90)
    ax.legend()
    fig.tight_layout()
    fig.savefig(
        path_save_plots / "datasets_log_scale.png",
        bbox_inches="tight",
    )
    plt.show()
    plt.close(fig)

    # ----- print some statistics ---------

    print(f"total number of time series: {sum(lengths)}")
    print(f"total number of time observations: {sum(n_total)}")

    f_size = sum([dataset_stats[d]["train_file_size"] for d in datasets])
    print(f"training set memory size: {f_size / float(1<<30)} GB")

    prediction_lengths = np.asarray(
        [dataset_stats[d]["prediction_length"] for d in datasets]
    )
    print(f"max prediction length is {np.max(prediction_lengths)}")


if __name__ == "__main__":
    main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional

from pandas.tseries.frequencies import to_offset


def lags_for_fourier_time_features_from_frequency(
    freq_str: str, num_lags: Optional[int] = None
) -> List[int]:
    offset = to_offset(freq_str)
    multiple, granularity = offset.n, offset.name

    if granularity in ("M", "ME"):
        lags = [[1, 12]]
    elif granularity == "D":
        lags = [[1, 7, 14]]
    elif granularity == "B":
        lags = [[1, 2]]
    elif granularity in ("H", "h"):
        lags = [[1, 24, 168]]
    elif granularity in ("T", "min"):
        lags = [[1, 4, 12, 24, 48]]
    else:
        lags = [[1]]

    # use less lags
    output_lags = list([int(lag) for sub_list in lags for lag in sub_list])
    output_lags = sorted(list(set(output_lags)))
    return output_lags[:num_lags]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional

import torch

from src.gluonts.dataset.field_names import FieldName
from src.gluonts.time_feature import TimeFeature
from src.gluonts.torch.model.predictor import PyTorchPredictor
from src.gluonts.torch.util import copy_parameters
from src.gluonts.model.predictor import Predictor
from src.gluonts.transform import (
    Transformation,
    Chain,
    InstanceSplitter,
    ExpectedNumInstanceSampler,
    ValidationSplitSampler,
    TestSplitSampler,
    RenameFields,
    AsNumpyArray,
    ExpandDimArray,
    AddObservedValuesIndicator,
    AddTimeFeatures,
    VstackFeatures,
    SetFieldIfNotPresent,
    TargetDimIndicator,
)

from pts import Trainer
from pts.feature import (
    fourier_time_features_from_frequency,
    lags_for_fourier_time_features_from_frequency,
)
from pts.model import PyTorchEstimator
from pts.model.utils import get_module_forward_input_names

from .time_grad_network import (
    TimeGradTrainingNetwork,
    TimeGradPredictionNetwork,
)


class TimeGradEstimator(PyTorchEstimator):
    def __init__(
        self,
        input_size: int,
        freq: str,
        prediction_length: int,
        target_dim: int,
        trainer: Trainer = Trainer(),
        context_length: Optional[int] = None,
        num_layers: int = 2,
        num_cells: int = 40,
        cell_type: str = "LSTM",
        num_parallel_samples: int = 100,
        dropout_rate: float = 0.1,
        cardinality: List[int] = [1],
        embedding_dimension: int = 5,
        conditioning_length: int = 100,
        diff_steps: int = 100,
        loss_type: str = "l2",
        beta_end=0.1,
        beta_schedule="linear",
        residual_layers=8,
        residual_channels=8,
        dilation_cycle_length=2,
        scaling: bool = True,
        pick_incomplete: bool = False,
        lags_seq: Optional[List[int]] = None,
        time_features: Optional[List[TimeFeature]] = None,
        **kwargs,
    ) -> None:
        super().__init__(trainer=trainer, **kwargs)

        self.freq = freq
        self.context_length = (
            context_length if context_length is not None else prediction_length
        )

        self.input_size = input_size
        self.prediction_length = prediction_length
        self.target_dim = target_dim
        self.num_layers = num_layers
        self.num_cells = num_cells
        self.cell_type = cell_type
        self.num_parallel_samples = num_parallel_samples
        self.dropout_rate = dropout_rate
        self.cardinality = cardinality
        self.embedding_dimension = embedding_dimension

        self.conditioning_length = conditioning_length
        self.diff_steps = diff_steps
        self.loss_type = loss_type
        self.beta_end = beta_end
        self.beta_schedule = beta_schedule
        self.residual_layers = residual_layers
        self.residual_channels = residual_channels
        self.dilation_cycle_length = dilation_cycle_length

        self.lags_seq = (
            lags_seq
            if lags_seq is not None
            else lags_for_fourier_time_features_from_frequency(freq_str=freq)
        )

        self.time_features = (
            time_features
            if time_features is not None
            else fourier_time_features_from_frequency(self.freq)
        )

        self.history_length = self.context_length + max(self.lags_seq)
        self.pick_incomplete = pick_incomplete
        self.scaling = scaling

        self.train_sampler = ExpectedNumInstanceSampler(
            num_instances=1.0,
            min_past=0 if pick_incomplete else self.history_length,
            min_future=prediction_length,
        )

        self.validation_sampler = ValidationSplitSampler(
            min_past=0 if pick_incomplete else self.history_length,
            min_future=prediction_length,
        )

    def create_transformation(self) -> Transformation:
        return Chain(
            [
                AsNumpyArray(
                    field=FieldName.TARGET,
                    expected_ndim=2,
                ),
                # maps the target to (1, T)
                # if the target data is uni dimensional
                ExpandDimArray(
                    field=FieldName.TARGET,
                    axis=None,
                ),
                AddObservedValuesIndicator(
                    target_field=FieldName.TARGET,
                    output_field=FieldName.OBSERVED_VALUES,
                ),
                AddTimeFeatures(
                    start_field=FieldName.START,
                    target_field=FieldName.TARGET,
                    output_field=FieldName.FEAT_TIME,
                    time_features=self.time_features,
                    pred_length=self.prediction_length,
                ),
                VstackFeatures(
                    output_field=FieldName.FEAT_TIME,
                    input_fields=[FieldName.FEAT_TIME],
                ),
                SetFieldIfNotPresent(
                    field=FieldName.FEAT_STATIC_CAT, value=[0]
                ),
                TargetDimIndicator(
                    field_name="target_dimension_indicator",
                    target_field=FieldName.TARGET,
                ),
                AsNumpyArray(field=FieldName.FEAT_STATIC_CAT, expected_ndim=1),
            ]
        )

    def create_instance_splitter(self, mode: str):
        assert mode in ["training", "validation", "test"]

        instance_sampler = {
            "training": self.train_sampler,
            "validation": self.validation_sampler,
            "test": TestSplitSampler(),
        }[mode]

        return InstanceSplitter(
            target_field=FieldName.TARGET,
            is_pad_field=FieldName.IS_PAD,
            start_field=FieldName.START,
            forecast_start_field=FieldName.FORECAST_START,
            instance_sampler=instance_sampler,
            past_length=self.history_length,
            future_length=self.prediction_length,
            time_series_fields=[
                FieldName.FEAT_TIME,
                FieldName.OBSERVED_VALUES,
            ],
        ) + (
            RenameFields(
                {
                    f"past_{FieldName.TARGET}": f"past_{FieldName.TARGET}_cdf",
                    f"future_{FieldName.TARGET}": f"future_{FieldName.TARGET}_cdf",
                }
            )
        )

    def create_training_network(
        self, device: torch.device
    ) -> TimeGradTrainingNetwork:
        return TimeGradTrainingNetwork(
            input_size=self.input_size,
            target_dim=self.target_dim,
            num_layers=self.num_layers,
            num_cells=self.num_cells,
            cell_type=self.cell_type,
            history_length=self.history_length,
            context_length=self.context_length,
            prediction_length=self.prediction_length,
            dropout_rate=self.dropout_rate,
            cardinality=self.cardinality,
            embedding_dimension=self.embedding_dimension,
            diff_steps=self.diff_steps,
            loss_type=self.loss_type,
            beta_end=self.beta_end,
            beta_schedule=self.beta_schedule,
            residual_layers=self.residual_layers,
            residual_channels=self.residual_channels,
            dilation_cycle_length=self.dilation_cycle_length,
            lags_seq=self.lags_seq,
            scaling=self.scaling,
            conditioning_length=self.conditioning_length,
        ).to(device)

    def create_predictor(
        self,
        transformation: Transformation,
        trained_network: TimeGradTrainingNetwork,
        device: torch.device,
    ) -> Predictor:
        prediction_network = TimeGradPredictionNetwork(
            input_size=self.input_size,
            target_dim=self.target_dim,
            num_layers=self.num_layers,
            num_cells=self.num_cells,
            cell_type=self.cell_type,
            history_length=self.history_length,
            context_length=self.context_length,
            prediction_length=self.prediction_length,
            dropout_rate=self.dropout_rate,
            cardinality=self.cardinality,
            embedding_dimension=self.embedding_dimension,
            diff_steps=self.diff_steps,
            loss_type=self.loss_type,
            beta_end=self.beta_end,
            beta_schedule=self.beta_schedule,
            residual_layers=self.residual_layers,
            residual_channels=self.residual_channels,
            dilation_cycle_length=self.dilation_cycle_length,
            lags_seq=self.lags_seq,
            scaling=self.scaling,
            conditioning_length=self.conditioning_length,
            num_parallel_samples=self.num_parallel_samples,
        ).to(device)

        copy_parameters(trained_network, prediction_network)
        input_names = get_module_forward_input_names(prediction_network)
        prediction_splitter = self.create_instance_splitter("test")

        return PyTorchPredictor(
            input_transform=transformation + prediction_splitter,
            input_names=input_names,
            prediction_net=prediction_network,
            batch_size=self.trainer.batch_size,
            freq=self.freq,
            prediction_length=self.prediction_length,
            device=device,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from copy import deepcopy
from functools import partial
from typing import List, Tuple, Optional

from mxnet.gluon import HybridBlock
import numpy as np
import pandas as pd

from src.gluonts.core.component import Type, validated
from src.gluonts.dataset.common import DataEntry, Dataset
from src.gluonts.dataset.field_names import FieldName
from src.gluonts.dataset.loader import (
    DataLoader,
    TrainDataLoader,
    ValidationDataLoader,
)
from src.gluonts.model.predictor import Predictor
from src.gluonts.mx.distribution import GaussianOutput
from src.gluonts.mx.model.estimator import GluonEstimator
from src.gluonts.mx.trainer import Trainer
from src.gluonts.mx.util import get_hybrid_forward_input_names
from src.gluonts.mx.util import copy_parameters
from src.gluonts.time_feature import (
    get_lags_for_frequency,
    time_features_from_frequency_str,
)
from src.gluonts.transform import (
    SelectFields,
    SimpleTransformation,
    Transformation,
    MissingValueImputation,
    RollingMeanValueImputation,
)
from src.gluonts.nursery.temporal_hierarchical_forecasting.model.cop_deepar.gluonts_fixes import (
    batchify_with_dict,
    DeepAREstimatorForCOP,
    LOSS_FUNCTIONS,
    RepresentableBlockPredictorBatchifyWithDict,
)
from src.gluonts.nursery.temporal_hierarchical_forecasting.utils.common import (
    TEMPORAL_HIERARCHIES,
)
from src.gluonts.nursery.temporal_hierarchical_forecasting.model.cop_deepar._network import (
    COPDeepARTrainingNetwork,
    COPDeepARPredictionNetwork,
)


class AddTimeFeaturesAtAggregateLevels(SimpleTransformation):
    @validated()
    def __init__(
        self,
        agg_multiples: List[int],
        agg_estimators: List[GluonEstimator],
    ):
        self.agg_multiples = agg_multiples
        self.agg_estimators = agg_estimators

    def transform(self, data: DataEntry) -> DataEntry:
        past_length_bottom_ts = data[f"past_{FieldName.FEAT_TIME}"].shape[1]

        agg_features_dict = {}
        for i, (agg_multiple, agg_estimator) in enumerate(
            zip(self.agg_multiples, self.agg_estimators)
        ):
            (
                past_time_feat_agg,
                future_time_feat_agg,
            ) = self._get_time_features_agg_level(
                forecast_start=data[FieldName.FORECAST_START],
                past_length_bottom_ts=past_length_bottom_ts,
                agg_estimator=agg_estimator,
            )
            agg_features_dict[f"level_{i}"] = {
                "past_time_feat_agg": past_time_feat_agg,
                "future_time_feat_agg": future_time_feat_agg,
            }
            data.update({"agg_features_dict": agg_features_dict})
        return data

    @staticmethod
    def _get_time_features_agg_level(
        forecast_start: pd.Period,
        past_length_bottom_ts: int,
        agg_estimator: GluonEstimator,
    ) -> Tuple[np.ndarray, np.ndarray]:
        start = forecast_start - past_length_bottom_ts

        freq = agg_estimator.freq
        num_periods = (
            agg_estimator.history_length + agg_estimator.prediction_length
        )
        full_date_range = pd.period_range(
            start,
            periods=num_periods,
            freq=freq,
        )

        # shape: (T, num_features)
        full_time_feat = np.array(
            [
                feat_map(full_date_range)
                for feat_map in time_features_from_frequency_str(freq)
            ]
        ).T

        age_feature = np.log10(
            2.0 + np.arange(num_periods, dtype=agg_estimator.dtype)
        ).reshape((num_periods, 1))
        full_time_feat = np.hstack((full_time_feat, age_feature))

        past_time_feat = full_time_feat[: agg_estimator.history_length, :]
        future_time_feat = full_time_feat[agg_estimator.history_length :, :]

        return past_time_feat, future_time_feat


class COPDeepAREstimator(GluonEstimator):
    """
    Construct a COP estimator for temporal hierarchies.

    Parameters
    ----------
    freq
    prediction_length
    base_estimator_name
        Currently only `DeepAREstimator`
    base_estimator_hps
    do_reconciliation
        Flag to indicate if the samples at different aggregated levels should be reconciled. Useful for ablation study.
        Set this to False and check just learning multiple related tasks at the same time affects the accuracy at the
        bottom level; this is without doing any reconciliation.
    loss_function
        Training loss function to be used. Currently one of `crps_univariate` and `mse`.
    distr_output
    trainer
    lag_ub
    num_samples_for_loss
    dtype
    """

    @validated()
    def __init__(
        self,
        freq: str,
        prediction_length: int,
        base_estimator_name: str,
        base_estimator_hps: dict,
        use_gnn: bool = True,
        use_mlp: bool = True,
        adj_mat_option: str = "hierarchical",
        non_negative: bool = False,
        do_reconciliation: bool = True,
        loss_function: str = "crps_univariate",
        warmstart_epoch_frac: float = 0.5,
        trainer: Trainer = Trainer(hybridize=False),
        lag_ub: int = 1200,
        num_samples_for_loss: int = 200,
        return_forecasts_at_all_levels: bool = False,
        naive_reconciliation: bool = False,
        dtype: Type = np.float32,
        impute_missing_values: bool = False,
        imputation_method: Optional[MissingValueImputation] = None,
        num_imputation_samples: int = 1,
    ) -> None:
        super().__init__(trainer=trainer, dtype=dtype)

        self.use_gnn = use_gnn
        self.use_mlp = use_mlp
        self.adj_mat_option = adj_mat_option
        self.do_reconciliation = do_reconciliation
        self.non_negative = non_negative
        self.warmstart_epoch_frac = warmstart_epoch_frac
        self.lag_ub = lag_ub
        self.num_samples_for_loss = num_samples_for_loss
        self.return_forecasts_at_all_levels = return_forecasts_at_all_levels
        self.naive_reconciliation = naive_reconciliation

        assert loss_function in LOSS_FUNCTIONS
        self.loss_function = loss_function

        self.context_length = base_estimator_hps.pop(
            "context_length", prediction_length
        )
        self.prediction_length = base_estimator_hps.pop(
            "prediction_length", prediction_length
        )
        self.freq = base_estimator_hps.pop("freq", freq)
        self.dtype = dtype

        self.temporal_hierarchy = TEMPORAL_HIERARCHIES[self.freq]
        self.base_estimator_type = eval(base_estimator_name)

        assert self.base_estimator_type == DeepAREstimatorForCOP

        base_estimator_hps.setdefault("distr_output", GaussianOutput())

        print(f"Distribution output: {base_estimator_hps['distr_output']}")

        base_estimator_hps.setdefault(
            "impute_missing_values", impute_missing_values
        )

        base_estimator_hps.setdefault("imputation_method", imputation_method)

        base_estimator_hps.setdefault(
            "num_imputation_samples", num_imputation_samples
        )

        self.estimators = []
        for agg_multiple, freq_str in zip(
            self.temporal_hierarchy.agg_multiples,
            self.temporal_hierarchy.freq_strs,
        ):
            base_estimator_hps_agg = deepcopy(base_estimator_hps)
            # Following is a hack because gluonts does not allow setting lag_ub from outside!
            lags_seq = get_lags_for_frequency(
                freq_str=freq_str,
                lag_ub=lag_ub // agg_multiple,
            )

            # Remove lags that will not be available for reconciliation during inference.
            num_nodes = self.temporal_hierarchy.num_leaves // agg_multiple
            lags_seq = [lag for lag in lags_seq if lag >= num_nodes]

            # adapt window_length if RollingMeanValueImputation is used
            if isinstance(imputation_method, RollingMeanValueImputation):
                base_estimator_hps_agg["imputation_method"] = (
                    RollingMeanValueImputation(
                        window_size=imputation_method.window_size
                        // agg_multiple
                    )
                )

            # Hack to enforce correct serialization of lags_seq and history length
            # (only works when set in constructor).
            if agg_multiple != 1:
                estimator = self.base_estimator_type(
                    freq=freq_str,
                    context_length=self.context_length // agg_multiple,
                    lags_seq=lags_seq,
                    prediction_length=self.prediction_length // agg_multiple,
                    **base_estimator_hps_agg,
                )
            else:
                estimator = self.base_estimator_type(
                    freq=freq_str,
                    context_length=self.context_length // agg_multiple,
                    lags_seq=lags_seq,
                    prediction_length=self.prediction_length // agg_multiple,
                    **base_estimator_hps_agg,
                    # Load more data at the bottom level so that enough lags are available for the aggregated levels.
                    history_length=(
                        (self.context_length // agg_multiple) + max(lags_seq)
                    )
                    * 2,
                )

            self.estimators.append(estimator)

        self.agg_estimators = self.estimators[:-1]
        self.base_estimator = self.estimators[-1]

        self.agg_feature_adder = AddTimeFeaturesAtAggregateLevels(
            agg_multiples=self.temporal_hierarchy.agg_multiples[:-1],
            agg_estimators=self.agg_estimators,
        )

    def create_transformation(self) -> Transformation:
        # We use the base estimator to create the data transformation.
        return self.base_estimator.create_transformation()

    def create_training_data_loader(
        self,
        data: Dataset,
        **kwargs,
    ) -> DataLoader:
        input_names = get_hybrid_forward_input_names(COPDeepARTrainingNetwork)
        instance_splitter = self.base_estimator._create_instance_splitter(
            "training"
        )

        return TrainDataLoader(
            dataset=data,
            transform=instance_splitter
            + self.agg_feature_adder
            + SelectFields(input_names),
            batch_size=self.batch_size,
            stack_fn=partial(
                batchify_with_dict, ctx=self.trainer.ctx, dtype=self.dtype
            ),
            **kwargs,
        )

    def create_validation_data_loader(
        self,
        data: Dataset,
        **kwargs,
    ) -> DataLoader:
        input_names = get_hybrid_forward_input_names(COPDeepARTrainingNetwork)
        instance_splitter = self.base_estimator._create_instance_splitter(
            "validation"
        )
        return ValidationDataLoader(
            dataset=data,
            transform=instance_splitter
            + self.agg_feature_adder
            + SelectFields(input_names),
            batch_size=self.batch_size,
            stack_fn=partial(
                batchify_with_dict, ctx=self.trainer.ctx, dtype=self.dtype
            ),
        )

    def create_training_network(self) -> COPDeepARTrainingNetwork:
        return COPDeepARTrainingNetwork(
            estimators=self.estimators,
            prediction_length=self.prediction_length,
            temporal_hierarchy=self.temporal_hierarchy,
            num_batches_per_epoch=self.trainer.num_batches_per_epoch,
            epochs=self.trainer.epochs,
            warmstart_epoch_frac=self.warmstart_epoch_frac,
            num_samples_for_loss=self.num_samples_for_loss,
            use_gnn=self.use_gnn,
            use_mlp=self.use_mlp,
            adj_mat_option=self.adj_mat_option,
            do_reconciliation=self.do_reconciliation,
            non_negative=self.non_negative,
            naive_reconciliation=self.naive_reconciliation,
            loss_function=self.loss_function,
            dtype=self.dtype,
        )

    def create_predictor(
        self, transformation: Transformation, trained_network: HybridBlock
    ) -> Predictor:
        prediction_splitter = self.base_estimator._create_instance_splitter(
            "test"
        )

        prediction_network = COPDeepARPredictionNetwork(
            estimators=self.estimators,
            prediction_length=self.prediction_length,
            temporal_hierarchy=self.temporal_hierarchy,
            use_gnn=self.use_gnn,
            use_mlp=self.use_mlp,
            adj_mat_option=self.adj_mat_option,
            do_reconciliation=self.do_reconciliation,
            non_negative=self.non_negative,
            naive_reconciliation=self.naive_reconciliation,
            return_forecasts_at_all_levels=self.return_forecasts_at_all_levels,
            num_parallel_samples=self.base_estimator.num_parallel_samples,
            dtype=self.dtype,
        )

        copy_parameters(trained_network, prediction_network)
        return RepresentableBlockPredictorBatchifyWithDict(
            input_transform=transformation
            + prediction_splitter
            + self.agg_feature_adder,
            prediction_net=prediction_network,
            batch_size=self.batch_size,
            prediction_length=self.prediction_length,
            ctx=self.trainer.ctx,
            dtype=self.dtype,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Tuple, Optional

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from pts.modules import MeanScaler, NOPScaler


class LSTNetBase(nn.Module):
    def __init__(
        self,
        num_series: int,
        channels: int,
        kernel_size: int,
        rnn_cell_type: str,
        rnn_num_cells: int,
        skip_rnn_cell_type: str,
        skip_rnn_num_cells: int,
        skip_size: int,
        ar_window: int,
        context_length: int,
        horizon: Optional[int],
        prediction_length: Optional[int],
        dropout_rate: float,
        output_activation: Optional[str],
        scaling: bool,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.num_series = num_series
        self.channels = channels
        assert (
            channels % skip_size == 0
        ), "number of conv1d `channels` must be divisible by the `skip_size`"
        self.skip_size = skip_size
        assert (
            ar_window > 0
        ), "auto-regressive window must be a positive integer"
        self.ar_window = ar_window
        assert not ((horizon is None)) == (
            prediction_length is None
        ), "Exactly one of `horizon` and `prediction_length` must be set at a time"
        assert (
            horizon is None or horizon > 0
        ), "`horizon` must be greater than zero"
        assert (
            prediction_length is None or prediction_length > 0
        ), "`prediction_length` must be greater than zero"
        self.prediction_length = prediction_length
        self.horizon = horizon
        assert context_length > 0, "`context_length` must be greater than zero"
        self.context_length = context_length
        if output_activation is not None:
            assert output_activation in [
                "sigmoid",
                "tanh",
            ], "`output_activation` must be either 'sigmiod' or 'tanh' "
        self.output_activation = output_activation
        assert rnn_cell_type in [
            "GRU",
            "LSTM",
        ], "`rnn_cell_type` must be either 'GRU' or 'LSTM' "
        assert skip_rnn_cell_type in [
            "GRU",
            "LSTM",
        ], "`skip_rnn_cell_type` must be either 'GRU' or 'LSTM' "

        conv_out = context_length - kernel_size
        self.conv_skip = conv_out // skip_size
        assert self.conv_skip > 0, (
            "conv1d output size must be greater than or equal to `skip_size`\n"
            "Choose a smaller `kernel_size` or bigger `context_length`"
        )

        self.cnn = nn.Conv2d(
            in_channels=1,
            out_channels=channels,
            kernel_size=(num_series, kernel_size),
        )

        self.dropout = nn.Dropout(p=dropout_rate)

        rnn = {"LSTM": nn.LSTM, "GRU": nn.GRU}[rnn_cell_type]
        self.rnn = rnn(
            input_size=channels,
            hidden_size=rnn_num_cells,
            # dropout=dropout_rate,
        )

        skip_rnn = {"LSTM": nn.LSTM, "GRU": nn.GRU}[skip_rnn_cell_type]
        self.skip_rnn_num_cells = skip_rnn_num_cells
        self.skip_rnn = skip_rnn(
            input_size=channels,
            hidden_size=skip_rnn_num_cells,
            # dropout=dropout_rate,
        )

        self.fc = nn.Linear(
            rnn_num_cells + skip_size * skip_rnn_num_cells, num_series
        )

        if self.horizon:
            self.ar_fc = nn.Linear(ar_window, 1)
        else:
            self.ar_fc = nn.Linear(ar_window, prediction_length)

        if scaling:
            self.scaler = MeanScaler(keepdim=True, time_first=False)
        else:
            self.scaler = NOPScaler(keepdim=True, time_first=False)

    def forward(
        self, past_target: torch.Tensor, past_observed_values: torch.Tensor
    ) -> torch.Tensor:
        scaled_past_target, scale = self.scaler(
            past_target[..., -self.context_length :],  # [B, C, T]
            past_observed_values[..., -self.context_length :],  # [B, C, T]
        )

        # CNN
        c = F.relu(self.cnn(scaled_past_target.unsqueeze(1)))
        c = self.dropout(c)
        c = c.squeeze(2)  # [B, C, T]

        # RNN
        r = c.permute(2, 0, 1)  # [F (T), B, C]
        _, r = self.rnn(r)  # [1, B, H]
        r = self.dropout(r.squeeze(0))  # [B, H]

        # Skip-RNN
        skip_c = c[..., -self.conv_skip * self.skip_size :]
        skip_c = skip_c.reshape(
            -1, self.channels, self.conv_skip, self.skip_size
        )
        skip_c = skip_c.permute(2, 0, 3, 1)
        skip_c = skip_c.reshape((self.conv_skip, -1, self.channels))
        _, skip_c = self.skip_rnn(skip_c)
        skip_c = skip_c.reshape((-1, self.skip_size * self.skip_rnn_num_cells))
        skip_c = self.dropout(skip_c)

        res = self.fc(torch.cat((r, skip_c), 1)).unsqueeze(-1)

        # Highway
        ar_x = scaled_past_target[..., -self.ar_window :]
        ar_x = ar_x.reshape(-1, self.ar_window)

        ar_x = self.ar_fc(ar_x)
        if self.horizon:
            ar_x = ar_x.reshape(-1, self.num_series, 1)
        else:
            ar_x = ar_x.reshape(-1, self.num_series, self.prediction_length)
        out = res + ar_x

        if self.output_activation is None:
            return out, scale

        return (
            (
                torch.sigmoid(out)
                if self.output_activation == "sigmoid"
                else torch.tanh(out)
            ),
            scale,
        )


class LSTNetTrain(LSTNetBase):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.loss_fn = nn.L1Loss()

    def forward(
        self,
        past_target: torch.Tensor,
        past_observed_values: torch.Tensor,
        future_target: torch.Tensor,
    ) -> torch.Tensor:
        ret, scale = super().forward(past_target, past_observed_values)

        if self.horizon:
            future_target = future_target[..., -1:]

        loss = self.loss_fn(ret * scale, future_target)
        return loss


class LSTNetPredict(LSTNetBase):
    def forward(
        self, past_target: torch.Tensor, past_observed_values: torch.Tensor
    ) -> torch.Tensor:
        ret, scale = super().forward(past_target, past_observed_values)
        ret = (ret * scale).permute(0, 2, 1)

        return ret.unsqueeze(1)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.
#
# Code from:
# https://github.com/White-Link/UnsupervisedScalableRepresentationLearningTimeSeries/blob/master/networks/causal_cnn.py
# which is based on:
# https://github.com/locuslab/TCN/blob/master/TCN/tcn.py, originally created
# with the following license.
#
# MIT License
# Copyright (c) 2018 CMU Locus Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import torch
import torch.nn.functional as F


class SqueezeChannels(torch.nn.Module):
    """
    Squeezes, in a three-dimensional tensor, the third dimension.
    """

    def __init__(self):
        super(SqueezeChannels, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x.squeeze(2)


class CausalConvolution(torch.nn.Module):
    """
    A single causal convolution applies the causal convolution itself, weight
    normalization, and an activation function.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        dilation: int,
    ):
        """
        Args:
            in_channels: The number of input channels.
            out_channels: The number of output channels.
            kernel_size: The kernel size to use for the convolution.
            dilation: The dilation parameter to use for the causal convolution.
        """
        super().__init__()

        conv = torch.nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            dilation=dilation,
        )
        self.padding = (kernel_size - 1) * dilation
        self.normalized_conv = torch.nn.utils.weight_norm(conv)
        self.activation = torch.nn.LeakyReLU()

    def forward(self, sequences: torch.Tensor) -> torch.Tensor:
        """
        Computes the causal convolution for the provided sequence. Inputs are
        padded such that the output sequence length.

        Args:
            sequences: Tensor of shape `[batch_size, in_channels, sequence_length]`.

        Returns:
            Tensor of shape `[batch_size, out_channels, sequence_length]`.
        """
        # Pad sequences on the left only
        padded_sequences = F.pad(sequences, pad=(self.padding, 0))
        # After running the convolution, the length of the sequence equals the input's
        out = self.normalized_conv(padded_sequences)
        return self.activation(out)


class CausalConvolutionBlock(torch.nn.Module):
    """
    Causal convolution block, composed sequentially of two causal convolutions
    (with leaky ReLU activation functions), and a parallel residual connection.
    Takes as input a three-dimensional tensor (`B`, `C`, `L`) where `B` is the
    batch size, `C` is the number of input channels (= number of dimensions
    of the multivariate time series), and `L` is the length of
    the input. Outputs a three-dimensional tensor (`B`, `C`, `L`).
    Parameters
    ----------
    in_channels : Number of input channels.
    out_channels : Number of output channels.
    kernel_size : Kernel size of the applied non-residual convolutions.
    dilation : Dilation parameter of non-residual convolutions.
    final : Disables, if True, the last activation function.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        dilation: int,
        final: bool = False,
    ):
        super(CausalConvolutionBlock, self).__init__()

        # Computes left padding so that the applied convolutions are causal
        padding = (kernel_size - 1) * dilation

        # First causal convolution
        conv1 = CausalConvolution(
            in_channels,
            out_channels,
            kernel_size,
            dilation=dilation,
        )

        # Second causal convolution
        conv2 = CausalConvolution(
            out_channels,
            out_channels,
            kernel_size,
            dilation=dilation,
        )

        # Causal network
        self.causal = torch.nn.Sequential(conv1, conv2)

        # Residual connection
        self.upordownsample = (
            torch.nn.Conv1d(in_channels, out_channels, 1)
            if in_channels != out_channels
            else None
        )

        # Final activation function
        self.relu = torch.nn.LeakyReLU() if final else None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out_causal = self.causal(x)
        if self.upordownsample is None:
            res = x
        else:
            res = self.upordownsample(x)
        if self.relu is None:
            return out_causal + res
        else:
            return self.relu(out_causal + res)


class CausalCNN(torch.nn.Module):
    """
    Causal CNN, composed of a sequence of causal convolution blocks.
    Takes as input a three-dimensional tensor (`B`, `C`, `L`) where `B` is the
    batch size, `C` is the number of input channels, and `L` is the length of
    the input. Outputs a three-dimensional tensor (`B`, `C_out`, `L`).
    Parameters
    ----------
    in_channels : Number of input channels.
    channels : Number of channels processed inside the network and of output
           channels.
    depth : Depth of the network.
    out_channels : Number of output channels.
    kernel_size : Kernel size of the applied non-residual convolutions.
    """

    def __init__(
        self,
        in_channels: int,
        channels: int,
        depth: int,
        out_channels: int,
        kernel_size: int,
    ):
        super(CausalCNN, self).__init__()

        layers = []  # List of causal convolution blocks
        dilation_size = 1  # Initial dilation size

        for i in range(depth):
            in_channels_block = in_channels if i == 0 else channels
            layers += [
                CausalConvolutionBlock(
                    in_channels_block, channels, kernel_size, dilation_size
                )
            ]
            dilation_size *= 2  # Doubles the dilation size at each step

        # Last layer
        layers += [
            CausalConvolutionBlock(
                channels, out_channels, kernel_size, dilation_size
            )
        ]

        self.network = torch.nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.network(x)


class CausalCNNEncoder(torch.nn.Module):
    """
    Encoder of a time series using a causal CNN: the computed representation is
    the output of a fully connected layer applied to the output of an adaptive
    max pooling layer applied on top of the causal CNN, which reduces the
    length of the time series to a fixed size.
    Takes as input a three-dimensional tensor (`B`, `C`, `L`) where `B` is the
    batch size, `C` is the number of input channels, and `L` is the length of
    the input. Outputs a three-dimensional tensor (`B`, `C`).
    Parameters
    ----------
    in_channels : Number of input channels.
    channels : Number of channels manipulated in the causal CNN.
    depth : Depth of the causal CNN.
    reduced_size : Fixed length to which the output time series of the
           causal CNN is reduced.
    out_channels : Number of output channels.
    kernel_size : Kernel size of the applied non-residual convolutions.
    """

    def __init__(
        self,
        in_channels: int,
        channels: int,
        depth: int,
        out_channels: int,
        kernel_size: int,
    ):
        super(CausalCNNEncoder, self).__init__()
        causal_cnn = CausalCNN(
            in_channels, channels, depth, out_channels, kernel_size
        )
        self.network = torch.nn.Sequential(
            causal_cnn,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        u = self.network(x)
        return u
